#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import annotations
"""
phonon_band.py
Make phonopy band.conf directly from POSCAR using SeeK-path.
Also includes Gruneisen parameter calculation workflow.
"""

import math
import sys
import subprocess
import shutil
from concurrent.futures import ProcessPoolExecutor
import matplotlib.pyplot as plt
import numpy as np
import os
from pathlib import Path
import traceback
import tempfile
import yaml
from macer.defaults import DEFAULT_MODELS, _macer_root, _model_root, resolve_model_path
from macer.calculator.factory import get_calculator

_PB_WORKER_CALC = None


def _init_pb_force_worker(ff_name, device, modal, model_path):
    global _PB_WORKER_CALC
    calc_kwargs = {"device": device, "modal": modal}
    if ff_name == "mace":
        calc_kwargs["model_paths"] = [model_path]
    else:
        calc_kwargs["model_path"] = model_path
    _PB_WORKER_CALC = get_calculator(ff_name, **calc_kwargs)


def _pb_force_worker(task):
    from ase import Atoms
    symbols, cell, scaled_positions = task
    atoms = Atoms(symbols=symbols, cell=cell, scaled_positions=scaled_positions, pbc=True)
    atoms.calc = _PB_WORKER_CALC
    return atoms.get_forces()


def _pb_resolve_model_path(ff: str, model_path: Path | None) -> str | None:
    """Resolve model path: prefer user-provided, else DEFAULT_MODELS."""
    if model_path:
        return resolve_model_path(str(model_path))
    default_model_name = DEFAULT_MODELS.get(ff)
    if default_model_name:
        return resolve_model_path(default_model_name)
    return None


def _get_strain_scaling_factor(B_GPa: float) -> float:
    """
    Calculates a scaling factor for strain based on the bulk modulus.
    The function is an exponential decay of the form f(B) = 2**(1.4 - B/50).
    """
    if B_GPa <= 0:
        return 1.0
    return 2**(1.4 - B_GPa / 50.0)


def _pb_generate_displacements(
    input_path: Path,
    min_length: float,
    displacement_distance: float,
    is_plusminus: bool,
    is_diagonal: bool,
    tolerance_phonopy: float,
    fix_axis: list[str] | None,
    dim_override: str | None,
):
    """Helper function to generate displacements with phonopy."""
    from phonopy.interface.vasp import read_vasp, write_vasp
    from phonopy import Phonopy
    
    print("\n--- Step 1: Generating displacements with phonopy ---")
    unitcell = read_vasp(str(input_path))
    cell = unitcell.cell
    vector_lengths = [np.linalg.norm(v) for v in cell]

    if any(v == 0 for v in vector_lengths):
        raise RuntimeError("One of the lattice vectors has a length of zero.")

    supercell_matrix = None
    if dim_override:
        try:
            parts = [int(p) for p in dim_override.split()]
        except ValueError:
            sys.exit("Error: --dim must be followed by integers.")

        if len(parts) == 3:
            supercell_matrix = np.diag(parts)
            print(f"Using user-provided diagonal supercell DIM = {parts}")
        elif len(parts) == 9:
            supercell_matrix = np.array(parts).reshape(3, 3)
            print(f"Using user-provided supercell matrix:\n{supercell_matrix}")
        else:
            sys.exit("Error: --dim must be followed by 3 or 9 integers.")
    else:
        # Auto-determine from min_length
        scaling_factors_list = [math.ceil(min_length / v) if v > 0 else 1 for v in vector_lengths]
        supercell_matrix = np.diag(scaling_factors_list)
        print(f"Auto-determining supercell from --min-length. DIM = {scaling_factors_list}")

    # Handle fix_axis for diagonal matrices
    if fix_axis:
        if np.count_nonzero(supercell_matrix - np.diag(np.diagonal(supercell_matrix))) == 0:
             scaling_factors = np.diagonal(supercell_matrix).astype(int).tolist()
             axis_map = {'a': 0, 'b': 1, 'c': 2, 'x': 0, 'y': 1, 'z': 2}
             for axis_name in fix_axis:
                 idx = axis_map.get(axis_name.lower())
                 if idx is not None and 0 <= idx < len(scaling_factors):
                     scaling_factors[idx] = 1
             supercell_matrix = np.diag(scaling_factors)
             print(f"Applied --fix-axis, new DIM = {scaling_factors}")
        else:
             print("Warning: --fix-axis is ignored when a non-diagonal --dim matrix is provided.")

    if np.count_nonzero(supercell_matrix - np.diag(np.diagonal(supercell_matrix))) == 0:
        sf_print = np.diagonal(supercell_matrix).astype(int).tolist()
        print(f"Supercell (DIM): {sf_print[0]} {sf_print[1]} {sf_print[2]}")

    try:
        phonon = Phonopy(
            unitcell,
            supercell_matrix=supercell_matrix,
            primitive_matrix="auto",
            symprec=tolerance_phonopy
        )
    except RuntimeError as e:
        if "Determinant of primitive_matrix" in str(e):
            print(f"Warning: Phonopy auto-primitive determination failed ({e}).")
            print("Falling back to primitive_matrix=np.eye(3) (assuming input cell is primitive).")
            phonon = Phonopy(
                unitcell,
                supercell_matrix=supercell_matrix,
                primitive_matrix=np.eye(3),
                symprec=tolerance_phonopy
            )
        else:
            raise e
    phonon.generate_displacements(
        distance=displacement_distance,
        is_plusminus=is_plusminus,
        is_diagonal=is_diagonal,
    )
    phonon.save("phonopy_disp.yaml")
    print("Wrote phonopy_disp.yaml")
    write_vasp("SPOSCAR", phonon.supercell)
    print("Wrote SPOSCAR")
    return phonon


def _pb_produce_fc_and_save(phonon: Phonopy, force_sets: list):
    """Helper function to produce force constants and save files."""
    from phonopy.file_IO import write_FORCE_SETS, write_FORCE_CONSTANTS
    print("\n--- Step 3: Creating FORCE_SETS and FORCE_CONSTANTS ---")
    # Set forces into the phonon object
    phonon.forces = force_sets
    # Produce force constants from forces
    phonon.produce_force_constants()
    print("Force constants produced via API.")
    
    # Save files
    write_FORCE_SETS(phonon.dataset, filename="FORCE_SETS")
    print("FORCE_SETS created successfully.")
    write_FORCE_CONSTANTS(phonon.force_constants, filename="FORCE_CONSTANTS")
    print("FORCE_CONSTANTS created successfully.")


def _calculate_phonons_for_gruneisen(
    input_path: Path,
    min_length: float,
    displacement_distance: float,
    is_plusminus: bool,
    is_diagonal: bool,
    macer_ff: str,
    macer_model_path: Path | None,
    macer_device: str,
    macer_modal: str | None,
    tolerance_phonopy: float,
    fix_axis: list[str] | None,
    dim_override: str | None,
    calculator=None,
):
    """Helper function to calculate force constants for a given structure."""
    phonon = _pb_generate_displacements(
        input_path, min_length, displacement_distance, 
        is_plusminus, is_diagonal, tolerance_phonopy, fix_axis, dim_override
    )

    print("\n--- Step 2: Calculating forces with macer ---")
    if calculator is None:
        print(f"Device: {macer_device}")
        current_macer_model_path = resolve_model_path(str(macer_model_path)) if macer_model_path else None
        if current_macer_model_path is None:
            default_model_name = DEFAULT_MODELS.get(macer_ff)
            if default_model_name:
                current_macer_model_path = resolve_model_path(default_model_name)

        if macer_ff == "mace" and macer_device == "mps":
            print(
                "\nWARNING: The 'mps' device is often incompatible with 'mace-torch' due to an internal "
                "bug related to float64 data types. Falling back to 'cpu' for force calculations to prevent a crash. "
                "For a permanent solution and better performance, please upgrade your 'mace-torch' package.",
                file=sys.stderr,
            )
            macer_device = "cpu"

        print(f"Loading MLFF model ({macer_ff})...")
        calc_kwargs = {"device": macer_device, "modal": macer_modal}
        if macer_ff == "mace":
            calc_kwargs["model_paths"] = [current_macer_model_path]
        else:
            calc_kwargs["model_path"] = current_macer_model_path
        calculator = get_calculator(macer_ff, **calc_kwargs)

    force_sets = []
    
    from ase import Atoms
    for cell in tqdm(phonon.supercells_with_displacements, desc="Calculating forces"):
        ase_atoms = Atoms(
            symbols=cell.symbols,
            cell=cell.cell,
            scaled_positions=cell.scaled_positions,
            pbc=True
        )
        ase_atoms.calc = calculator
        forces = ase_atoms.get_forces()
        force_sets.append(forces)
    print("Force calculations completed successfully.")

    _pb_produce_fc_and_save(phonon, force_sets)
    
    return phonon


def _read_band_yaml_data(band_yaml_path):
    """Reads data from a band.yaml file."""
    import yaml
    with open(band_yaml_path, 'r') as f:
        data = yaml.load(f, Loader=yaml.CLoader)

    frequencies = []
    distances = []
    qpoints = []
    for v in data["phonon"]:
        frequencies.append([f["frequency"] for f in v["band"]])
        distances.append(v["distance"])
        qpoints.append(v["q-position"])
    
    distances = np.array(distances)
    frequencies = np.array(frequencies)
    qpoints = np.array(qpoints)
    segment_nqpoint = data["segment_nqpoint"]
    labels = data.get("labels", [])
    
    # Reconstruct segmented q-points list for set_band_structure
    bands = []
    q_idx = 0
    for nq in segment_nqpoint:
        bands.append(qpoints[q_idx:q_idx + nq])
        q_idx += nq
        
    return distances, frequencies, bands, segment_nqpoint, labels


def _write_band_dat(band_yaml_path, dat_filename):
    """Writes a band.yaml file to a gnuplot-friendly .dat file."""
    distances, frequencies, _, segment_nqpoint, _ = _read_band_yaml_data(band_yaml_path)

    with open(dat_filename, 'w') as f:
        f.write("# q-distance, frequency\n")
        for i, freqs_band in enumerate(frequencies.T):  # Iterate over bands
            f.write(f"# mode {i + 1}\n")
            q_idx = 0
            for nq in segment_nqpoint:  # Iterate over segments
                for d, freq in zip(distances[q_idx:q_idx + nq], 
                                   freqs_band[q_idx:q_idx + nq]):
                    f.write(f"{d:12.8f} {freq:15.8f}\n")
                q_idx += nq
                f.write("\n")  # Blank line between segments for gnuplot
            f.write("\n")  # Another blank line between bands
    
    print(f"Band structure data written to {dat_filename}")


def _write_gruneisen_dat(distances, frequencies, gruneisen_params, segment_nqpoint, dat_filename="gruneisen.dat"):
    """Write band structure data to a text file in gnuplot-friendly format."""
    num_bands = frequencies.shape[1]

    with open(dat_filename, 'w') as f:
        f.write("# q-distance, frequency, gruneisen\n")
        for j in range(num_bands):  # Loop over bands (modes)
            f.write(f"# mode {j + 1}\n")
            q_idx = 0
            for nq in segment_nqpoint: # Iterate over segments
                for k in range(nq): # Loop over q-points in segment
                    idx = q_idx + k
                    dist = distances[idx]
                    freq = frequencies[idx, j]
                    g = gruneisen_params[idx, j]
                    if freq < 0.05:
                        g = 0.0
                    f.write(f"{dist:12.8f} {freq:15.8f} {g:15.8f}\n")
                q_idx += nq
                f.write("\n") # Blank line between segments
            f.write("\n\n")  # Two blank lines for gnuplot index separation
            
    print(f"Gruneisen data written to {dat_filename}")


def _plot_gruneisen_from_dat(dat_filename, distances, segment_nqpoint, labels, gmin, gmax, filter_outliers_factor, output_filename="gruneisen_band.pdf"):
    """Plot phonon dispersion colored by Gruneisen parameter from a data file."""
    import matplotlib.pyplot as plt
    data_points = []
    try:
        with open(dat_filename, 'r') as f:
            lines = f.readlines()
        for line in lines:
            if line.strip() and not line.strip().startswith('#'):
                parts = line.split()
                if len(parts) == 3:
                    data_points.append([float(p) for p in parts])
    except (IOError, ValueError) as e:
        print(f"Error reading or parsing {dat_filename}: {e}")
        return None, None

    if not data_points:
        print(f"No data found in {dat_filename}")
        return None, None

    data = np.array(data_points)
    data = data[np.isfinite(data).all(axis=1)]

    x = data[:, 0]
    y = data[:, 1]
    g = data[:, 2]

    # Determine limits and filtering
    q1, q3 = np.percentile(g, [25, 75])
    iqr = q3 - q1
    f_factor = filter_outliers_factor if filter_outliers_factor is not None else 3.0
    auto_vmin, auto_vmax = q1 - f_factor * iqr, q3 + f_factor * iqr
    
    # Check for significant negative values (tolerance -0.05)
    has_significant_negative = g.min() < -0.05
    
    if gmin is None:
        if not has_significant_negative:
            # Case: Effectively positive -> Start from 0
            vmin = 0.0
        else:
            # Case: Significant negative -> Use auto limit
            vmin = float(math.floor(max(g.min(), auto_vmin)))
    else:
        vmin = gmin

    vmax = gmax if gmax is not None else float(math.ceil(min(g.max(), auto_vmax)))
    
    mask = (g >= vmin) & (g <= vmax)
    print(f"Plotting range: [{vmin:.2f}, {vmax:.2f}]. Points outside this range are hidden.")
    x_filt, y_filt, g_filt = x[mask], y[mask], g[mask]

    fig, ax = plt.subplots(figsize=(8, 6))
    plt.rcParams["pdf.fonttype"] = 42
    plt.rcParams["font.family"] = "serif"

    # Color range handling: use TwoSlopeNorm to keep 0 as white even if asymmetric
    from matplotlib.colors import TwoSlopeNorm, Normalize, LinearSegmentedColormap
    
    # Force range to include 0 to ensure white center
    vmin = min(vmin, 0.0)
    vmax = max(vmax, 0.0)
    
    # Default cmap
    cmap_to_use = 'bwr'
    
    if vmin == 0 and vmax == 0:
        # Trivial case: all 0
        norm = Normalize(vmin=-1, vmax=1)
        ticks = [0]
    elif vmin == 0:
        # Positive only: Truncate bwr to White->Red (0.5 to 1.0)
        original_cmap = plt.get_cmap('bwr')
        colors = original_cmap(np.linspace(0.5, 1.0, 128))
        cmap_to_use = LinearSegmentedColormap.from_list('bwr_upper', colors)
        
        norm = Normalize(vmin=0, vmax=vmax)
        n_side = 4
        ticks = np.linspace(0, vmax, n_side + 1)
    elif vmax == 0:
        # Negative only: Truncate bwr to Blue->White (0.0 to 0.5)
        original_cmap = plt.get_cmap('bwr')
        colors = original_cmap(np.linspace(0.0, 0.5, 128))
        cmap_to_use = LinearSegmentedColormap.from_list('bwr_lower', colors)
        
        norm = Normalize(vmin=vmin, vmax=0)
        n_side = 4
        ticks = np.linspace(vmin, 0, n_side + 1)
    else:
        # Crossing 0: TwoSlopeNorm for full dynamic range on both sides
        norm = TwoSlopeNorm(vmin=vmin, vcenter=0, vmax=vmax)
        n_side = 4
        
        ticks_neg = np.linspace(vmin, 0, n_side + 1) if vmin < 0 else []
        ticks_pos = np.linspace(0, vmax, n_side + 1) if vmax > 0 else []
        
        # Combine and ensure unique 0
        if len(ticks_neg) > 0 and len(ticks_pos) > 0:
             ticks = np.concatenate([ticks_neg, ticks_pos[1:]])
        elif len(ticks_neg) > 0:
             ticks = ticks_neg
        else:
             ticks = ticks_pos

    sc = ax.scatter(x_filt, y_filt, c=g_filt, cmap=cmap_to_use, s=5, norm=norm, alpha=0.8)
    cb = fig.colorbar(sc, ax=ax, ticks=ticks)
    cb.set_label('Grüneisen parameter')

    # Ticks and labels extraction
    tick_pos = [distances[0]]
    tick_labels_raw = [labels[0][0]]
    curr_idx = 0
    for i, nq in enumerate(segment_nqpoint):
        curr_idx += nq
        tick_pos.append(distances[curr_idx - 1])
        tick_labels_raw.append(labels[i][1])

    clean_labels = [l.replace("GAMMA", "$\\Gamma$").replace("GM", "$\\Gamma$") for l in tick_labels_raw]
    ax.set_xticks(tick_pos)
    ax.set_xticklabels(clean_labels)
    ax.set_xlim(distances[0], distances[-1])
    ax.set_ylabel('Frequency (THz)')
    
    for tp in tick_pos:
        ax.axvline(tp, color='k', linestyle='-', linewidth=0.5, alpha=0.5)
    ax.axhline(0, linestyle='-', color='k', linewidth=0.5)

    plt.tight_layout()
    plt.savefig(output_filename)
    plt.close()
    print(f"Gruneisen band structure plot saved to {output_filename}")
    
    return vmin, vmax


def _plot_mode_gruneisen_from_dat(dat_filename, distances, segment_nqpoint, labels, gmin, gmax, filter_outliers_factor, output_filename="gruneisen_mode_parameter.pdf"):
    """Plot Mode Grüneisen parameter vs Wave vector from a data file."""
    import matplotlib.pyplot as plt
    data_points = []
    try:
        with open(dat_filename, 'r') as f:
            lines = f.readlines()
        for line in lines:
            if line.strip() and not line.strip().startswith('#'):
                parts = line.split()
                if len(parts) == 3:
                    data_points.append([float(p) for p in parts])
    except (IOError, ValueError) as e:
        print(f"Error reading or parsing {dat_filename}: {e}")
        return

    if not data_points:
        return

    data = np.array(data_points)
    data = data[np.isfinite(data).all(axis=1)]

    x = data[:, 0]
    g = data[:, 2] 

    # Determine limits and filtering
    q1, q3 = np.percentile(g, [25, 75])
    iqr = q3 - q1
    f_factor = filter_outliers_factor if filter_outliers_factor is not None else 3.0
    auto_vmin, auto_vmax = q1 - f_factor * iqr, q3 + f_factor * iqr
    
    # Check for significant negative values (tolerance -0.05)
    has_significant_negative = g.min() < -0.05
    
    if gmin is None:
        if not has_significant_negative:
            # Case: Effectively positive -> Start from 0
            vmin = 0.0
        else:
            # Case: Significant negative -> Use auto limit
            vmin = float(math.floor(max(g.min(), auto_vmin)))
    else:
        vmin = gmin

    vmax = gmax if gmax is not None else float(math.ceil(min(g.max(), auto_vmax)))
    
    mask = (g >= vmin) & (g <= vmax)
    x_filt, g_filt = x[mask], g[mask]

    fig, ax = plt.subplots(figsize=(8, 6))
    plt.rcParams["pdf.fonttype"] = 42
    plt.rcParams["font.family"] = "serif"

    from matplotlib.colors import TwoSlopeNorm, Normalize, LinearSegmentedColormap
    
    # Force range to include 0 to ensure white center
    vmin = min(vmin, 0.0)
    vmax = max(vmax, 0.0)
    
    # Default cmap
    cmap_to_use = 'bwr'
    
    if vmin == 0 and vmax == 0:
        # Trivial case: all 0
        norm = Normalize(vmin=-1, vmax=1)
        ticks = [0]
    elif vmin == 0:
        # Positive only: Truncate bwr to White->Red (0.5 to 1.0)
        original_cmap = plt.get_cmap('bwr')
        colors = original_cmap(np.linspace(0.5, 1.0, 128))
        cmap_to_use = LinearSegmentedColormap.from_list('bwr_upper', colors)
        
        norm = Normalize(vmin=0, vmax=vmax)
        n_side = 4
        ticks = np.linspace(0, vmax, n_side + 1)
    elif vmax == 0:
        # Negative only: Truncate bwr to Blue->White (0.0 to 0.5)
        original_cmap = plt.get_cmap('bwr')
        colors = original_cmap(np.linspace(0.0, 0.5, 128))
        cmap_to_use = LinearSegmentedColormap.from_list('bwr_lower', colors)
        
        norm = Normalize(vmin=vmin, vmax=0)
        n_side = 4
        ticks = np.linspace(vmin, 0, n_side + 1)
    else:
        # Crossing 0: TwoSlopeNorm for full dynamic range on both sides
        norm = TwoSlopeNorm(vmin=vmin, vcenter=0, vmax=vmax)
        n_side = 4
        
        ticks_neg = np.linspace(vmin, 0, n_side + 1) if vmin < 0 else []
        ticks_pos = np.linspace(0, vmax, n_side + 1) if vmax > 0 else []
        
        # Combine and ensure unique 0
        if len(ticks_neg) > 0 and len(ticks_pos) > 0:
             ticks = np.concatenate([ticks_neg, ticks_pos[1:]])
        elif len(ticks_neg) > 0:
             ticks = ticks_neg
        else:
             ticks = ticks_pos

    if vmin < 0 < vmax:
        from matplotlib.scale import FuncScale
        def forward(y): return np.where(y > 0, y / vmax, y / abs(vmin))
        def inverse(y): return np.where(y > 0, y * vmax, y * abs(vmin))
        ax.set_yscale('function', functions=(forward, inverse))
        ax.set_yticks(ticks)
    else:
        # If not crossing, standard linear scale but set ticks
        ax.set_yticks(ticks)

    sc = ax.scatter(x_filt, g_filt, c=g_filt, cmap=cmap_to_use, s=5, norm=norm, alpha=0.8)
    cb = fig.colorbar(sc, ax=ax, ticks=ticks)
    cb.set_label('Grüneisen parameter')

    # Ticks and labels extraction
    tick_pos = [distances[0]]
    tick_labels_raw = [labels[0][0]]
    curr_idx = 0
    for i, nq in enumerate(segment_nqpoint):
        curr_idx += nq
        tick_pos.append(distances[curr_idx - 1])
        tick_labels_raw.append(labels[i][1])

    clean_labels = [l.replace("GAMMA", "$\\Gamma$").replace("GM", "$\\Gamma$") for l in tick_labels_raw]
    ax.set_xticks(tick_pos)
    ax.set_xticklabels(clean_labels)
    ax.set_xlim(distances[0], distances[-1])
    ax.set_ylim(vmin, vmax)
    
    ax.set_ylabel('Mode Grüneisen parameter')
    for tp in tick_pos:
        ax.axvline(tp, color='k', linestyle='-', linewidth=0.5, alpha=0.5)
    ax.axhline(0, linestyle=':', color='gray', linewidth=0.8, alpha=0.5)

    plt.tight_layout()
    plt.savefig(output_filename)
    plt.close()
    print(f"Mode Grüneisen parameter plot saved to {output_filename}")



def _run_gruneisen_on_top(
    phonons: list[Phonopy],
    strain: float,
    band_yaml_path: Path,
    output_prefix: str,
    mesh: list[int] = [11, 11, 11],
    meta_formula: str | None = None,
    meta_sg: str | None = None,
    meta_path: str | None = None,
):
    """Calculate Gruneisen parameters using pre-calculated phonon objects."""
    print("\n===== Gruneisen calculation: START ====")
    from phonopy import PhonopyGruneisen
    
    gruneisen = PhonopyGruneisen(phonons[0], phonons[1], phonons[2], delta_strain=strain)
    
    # 1. Band Path Calculation (Existing)
    print("  Calculating Grüneisen parameters along band path...")
    distances, frequencies, bands, segment_nqpoint, labels = _read_band_yaml_data(band_yaml_path)
    
    gruneisen.set_band_structure(bands)
    _, _, _, _, gruneisen_params_list = gruneisen.get_band_structure()

    # Flatten the list of lists into single arrays
    gruneisen_params = np.vstack(gruneisen_params_list)

    dat_filename = Path.cwd() / f"gruneisen_band-{output_prefix}.dat"
    _write_gruneisen_dat(distances, frequencies, gruneisen_params, segment_nqpoint, dat_filename)
    
    # 2. Full Mesh Calculation (New)
    print(f"  Calculating Grüneisen parameters on full mesh {mesh}...")
    gruneisen.set_mesh(mesh, is_gamma_center=True)
    
    mesh_qpoints, mesh_weights, mesh_freqs, _, mesh_gammas = gruneisen.get_mesh()
    
    full_dat_filename = Path.cwd() / f"gruneisen_full-{output_prefix}.dat"
    
    with open(full_dat_filename, 'w') as f:
        if meta_formula: f.write(f"# Formula: {meta_formula}\n")
        if meta_sg: f.write(f"# SpaceGroup: {meta_sg}\n")
        if meta_path: f.write(f"# Path: {meta_path}\n")
        f.write("# qx qy qz weight band_index frequency(THz) gruneisen_param\n")
        
        n_q, n_bands = mesh_freqs.shape
        for i in range(n_q):
            q = mesh_qpoints[i]
            w = mesh_weights[i]
            for j in range(n_bands):
                freq = mesh_freqs[i, j]
                gamma = mesh_gammas[i, j]
                f.write(f"{q[0]:10.6f} {q[1]:10.6f} {q[2]:10.6f} {w:6d} {j+1:4d} {freq:15.8f} {gamma:15.8f}\n")
    
    print(f"  Full grid Grüneisen data written to {full_dat_filename.name}")
    
    return dat_filename, distances, segment_nqpoint, labels


def run_macer_workflow(
    input_path: Path,
    min_length: float,
    displacement_distance: float,
    is_plusminus: bool,
    is_diagonal: bool,
    symprec_seekpath: float = 1e-5,
    gamma_label: str = "GM",
    macer_device: str = "cpu",
    macer_model_path: Path | None = None,
    model_info_str: str = "",
    yaml_path_arg: Path | None = None,
    out_path_arg: Path | None = None,
    dim_override: str | None = None,
    no_defaults_band_conf: bool = False,
    atom_names_override: str | None = None,
    rename_override: str | None = None,
    output_prefix: str | None = None,
    tolerance_sr: float = 0.01,
    tolerance_phonopy: float = 5e-3,
    macer_optimizer_name: str = "FIRE",
    fix_axis: list[str] | None = None,
    macer_ff: str = "mace",
    macer_modal: str | None = None,
    plot_gruneisen: bool = False,
    gruneisen_strain: float | None = None,
    select_axis: str | None = None,
    gmin: float | None = None,
    gmax: float | None = None,
    gruneisen_target_energy: float = 10.0,
    filter_outliers_factor: float | None = None,
    use_relax_unit: bool = False,
    initial_fmax: float = 0.005,
    initial_symprec: float = 1e-5,
    initial_isif: int = 3,
    show_irreps: bool = False,
    irreps_qpoint: list[float] | None = None,
    tolerance_irreps: float = 1e-5,
    write_arrow: bool = False,
    arrow_length: float = 1.7,
    arrow_min_cutoff: float = 0.1,
    arrow_qpoint_gamma: bool = False,
    arrow_qpoint: list[float] | None = None,
    mass_map: dict | None = None,
    output_dir_arg: str | None = None,
    plot_dos: bool = False,
    mesh: list[int] = [11, 11, 11],  # Default updated to 11 11 11 in CLI, reflected here
    with_eigenvectors: bool = False,
    read_fc_path: Path | None = None,
    skip_initial_relax: bool = False,
    manual_band_path: str | None = None,
    manual_labels: str | None = None,
    n_jobs: int = 1,
):
    import traceback # Ensure local availability
    # ... (rest of the function start) ...
        
    # Lazy imports for workflow
    import macer.externals.spglib_bundled as spglib
    from ase.io import read as ase_read
    from phonopy.file_IO import parse_FORCE_CONSTANTS
    from phonopy.interface.vasp import read_vasp
    from phonopy.phonon.band_structure import get_band_qpoints
    from macer.phonopy.band_path import generate_band_conf
    from macer.relaxation.bulk_modulus import get_bulk_modulus_and_volume
    from ase import Atoms
    from ase.io import write as ase_write
    from types import SimpleNamespace
    from tqdm import tqdm

    final_symmetrized_sg_symbol = "N/A"
    final_symmetrized_sg_number = None

    input_path = Path(input_path).resolve()
    if n_jobs < 1:
        raise ValueError("Error: --n-jobs must be >= 1.")
    caller_cwd = Path.cwd()
    initial_atoms = ase_read(input_path, format="vasp")

    # Normalize model path type (accept both str and Path from callers).
    if macer_model_path:
        macer_model_path = Path(macer_model_path)

    # Preserve user-provided local model paths across internal chdir() calls.
    if macer_model_path and not macer_model_path.is_absolute():
        caller_relative_model = (caller_cwd / macer_model_path).resolve()
        if caller_relative_model.exists():
            macer_model_path = caller_relative_model

    # Parse selected axes for selective Gruneisen strain scaling.
    selected_axes_str = select_axis
    if selected_axes_str:
        selected_axes = {axis.strip().lower() for axis in selected_axes_str.split(',')}
        valid_axes = {'a', 'b', 'c'}
        if not selected_axes.issubset(valid_axes):
            raise ValueError(
                f"Error: Invalid axis in --select-axis. Allowed values are 'a', 'b', 'c'. "
                f"You provided: {selected_axes_str}"
            )
    else:
        selected_axes = None

    # Determine output directory
    if output_dir_arg:
        output_dir = Path(output_dir_arg).resolve()
    else:
        base_output_dir = input_path.parent / f"phonon_band-{input_path.name}-mlff={macer_ff}"
        output_dir = base_output_dir
        if output_dir.exists():
            i = 1
            while output_dir.exists():
                output_dir = Path(f"{base_output_dir}-NEW{i:02d}")
                i += 1

    if not output_dir.exists():
        output_dir.mkdir(parents=True)
    
    print(f"Output directory: {output_dir}")

    # Copy input file to output directory and update input_path
    local_input_path = output_dir / input_path.name
    if input_path.resolve() != local_input_path.resolve():
        shutil.copy(input_path, local_input_path)
    os.chdir(output_dir)
    input_path = local_input_path

    if output_prefix is None:
        output_prefix = input_path.stem

    unitcell_poscar_path = Path.cwd() / f"{input_path.stem}-symmetrized"
    
    print(f"Input structure: {input_path}")
    print(f"Output directory: {Path.cwd()}")
    print(f"Displacement distance: {displacement_distance:g} \u00c5")

    # === Step 0: Initial Relaxation ===
    print("\n--- Step 0: Initial structure preparation ---")
    symmetrized_input_path = None

    if read_fc_path and skip_initial_relax:
        print(f"Skipping Step 0-3: Using provided FORCE_CONSTANTS from {read_fc_path}")
        symmetrized_input_path = output_dir / "POSCAR"
        if not symmetrized_input_path.exists():
            shutil.copy(local_input_path, symmetrized_input_path)
    elif use_relax_unit:
        print("Using iterative relaxation/symmetrization (macer_phonopy sr).")
        macer_ru_command = [
            sys.executable, '-m', 'macer.cli.phonopy_main', 'sr', '-p', str(input_path), '--output-prefix', input_path.stem,
            '--tolerance', str(tolerance_sr), '--ff', macer_ff,
        ]
        if macer_model_path:
            macer_ru_command.extend(['--model', str(macer_model_path)])
        else: # Logic to find default model if not provided
            default_model_name = DEFAULT_MODELS.get(macer_ff)
            if default_model_name:
                # This logic needs to be careful about FFS_USING_MODEL_NAME
                FFS_USING_MODEL_NAME = {"fairchem", "orb", "chgnet", "m3gnet"}
                if macer_ff in FFS_USING_MODEL_NAME:
                    macer_ru_command.extend(['--model', default_model_name])
                else:
                    macer_ru_command.extend(['--model', os.path.join(_model_root, default_model_name)])

        if macer_modal: macer_ru_command.extend(['--modal', macer_modal])
        if fix_axis: macer_ru_command.extend(['--fix-axis', ','.join(fix_axis)])
        
        print(f"Running command: {' '.join(macer_ru_command)}")
        try:
            subprocess.run(macer_ru_command, check=True, text=True)
            symmetrized_input_path = unitcell_poscar_path
        except subprocess.CalledProcessError as e:
            print("Error during macer_phonopy sr execution.")
            return
    else:
        if initial_isif == 0:
            print("ISIF=0 specified, skipping initial relaxation.")
            symmetrized_input_path = output_dir / input_path.name
            if not symmetrized_input_path.exists():
                shutil.copy(input_path, symmetrized_input_path)
        else:
            print(f"Using simple 'macer relax' (ISIF={initial_isif}).")
            copied_input_poscar_path = output_dir / input_path.name
            if not copied_input_poscar_path.exists(): # Avoid overwriting if input is in output dir
                 shutil.copy(input_path, copied_input_poscar_path)

            relaxed_poscar_name = f"CONTCAR-{copied_input_poscar_path.stem}"
            macer_relax_command = [
                sys.executable, '-m', 'macer.cli.main', 'relax', '-p', str(copied_input_poscar_path), '--isif', str(initial_isif),
                '--fmax', str(initial_fmax), '--device', macer_device,
                '--contcar', relaxed_poscar_name, '--no-pdf',
                '--symprec', str(initial_symprec)
            ]
            if macer_ff: macer_relax_command.extend(['--ff', macer_ff])
            if macer_model_path: macer_relax_command.extend(['--model', str(macer_model_path)])
            if macer_modal: macer_relax_command.extend(['--modal', macer_modal])
            
            print(f"Running command: {' '.join(macer_relax_command)}")
            try:
                subprocess.run(macer_relax_command, check=True, text=True)
                symmetrized_input_path = output_dir / relaxed_poscar_name
                if not symmetrized_input_path.exists():
                    raise FileNotFoundError(f"Relaxed CONTCAR not found at {symmetrized_input_path}")
            except (subprocess.CalledProcessError, FileNotFoundError) as e:
                print("Error during initial 'macer relax' execution:", e)
                return

    # --- Post-relaxation info gathering ---
    symmetrized_atoms = ase_read(str(symmetrized_input_path), format="vasp")
    cell_tuple = (symmetrized_atoms.get_cell(),
                  symmetrized_atoms.get_scaled_positions(),
                  symmetrized_atoms.get_atomic_numbers())
    dataset = spglib.get_symmetry_dataset(cell_tuple, symprec=tolerance_sr)
    if dataset:
        final_symmetrized_sg_number = dataset.get('number')
        final_symmetrized_sg_symbol = dataset.get('international', "N/A")
    print(f"  Space group of prepared structure ({symmetrized_input_path.name}): {final_symmetrized_sg_symbol} (No. {final_symmetrized_sg_number})")

    # === Step 1: Displacement Generation & Force Calculation ===
    base_phonon = None
    all_phonons = []

    if read_fc_path:
        print("\n--- Step 1-3: Initializing Phonopy with existing FORCE_CONSTANTS ---")
        unitcell = read_vasp(str(symmetrized_input_path))
        # Reuse supercell matrix detection logic or use provided dim_override
        # (For simplicity here, we assume detected supercell matches provided FC)
        dummy_ph = _pb_generate_displacements(
            input_path=symmetrized_input_path, min_length=min_length, displacement_distance=displacement_distance,
            is_plusminus=is_plusminus, is_diagonal=is_diagonal, tolerance_phonopy=tolerance_phonopy,
            fix_axis=fix_axis, dim_override=dim_override
        )
        base_phonon = dummy_ph
        base_phonon.force_constants = parse_FORCE_CONSTANTS(read_fc_path)
        all_phonons = [base_phonon]
        # Copy to local FORCE_CONSTANTS for consistency
        shutil.copy(read_fc_path, output_dir / "FORCE_CONSTANTS")
    else:
        # 1.1 Pre-load MLFF model
        print(f"\n--- Step 1: Loading MLFF model ({macer_ff}) ---")
        current_macer_model_path = resolve_model_path(str(macer_model_path)) if macer_model_path else None
        if current_macer_model_path is None:
            default_model_name = DEFAULT_MODELS.get(macer_ff)
            if default_model_name:
                current_macer_model_path = resolve_model_path(default_model_name)

        if macer_ff == "mace" and macer_device == "mps":
            print(
                "\nWARNING: The 'mps' device is often incompatible with 'mace-torch' due to an internal "
                "bug related to float64 data types. Falling back to 'cpu' for force calculations to prevent a crash. ",
                file=sys.stderr,
            )
            macer_device = "cpu"

        calc_kwargs = {"device": macer_device, "modal": macer_modal}
        if macer_ff == "mace":
            calc_kwargs["model_paths"] = [current_macer_model_path]
        else:
            calc_kwargs["model_path"] = current_macer_model_path
        workflow_calculator = get_calculator(macer_ff, **calc_kwargs)

        # 1.2 Grüneisen strain estimation (Optional)
        final_gruneisen_strain = None
        if plot_gruneisen:
            final_gruneisen_strain = gruneisen_strain
            log_messages = []
            if final_gruneisen_strain is None:
                print("\n--- Estimating optimal strain for Grüneisen from bulk modulus ---")
                log_messages.append("Grüneisen strain not provided, estimating from bulk modulus.")
                calc_args = SimpleNamespace(ff=macer_ff, model=current_macer_model_path, device=macer_device, modal=macer_modal)
                B_GPa, V0_per_atom = get_bulk_modulus_and_volume(initial_atoms, calc_args)
                if B_GPa and V0_per_atom:
                    try:
                        E_target_eV = gruneisen_target_energy / 1000.0
                        B_eV_per_A3 = B_GPa / 160.21766208
                        epsilon_V = math.sqrt(2 * E_target_eV / (B_eV_per_A3 * V0_per_atom))
                        scaling_factor = _get_strain_scaling_factor(B_GPa)
                        final_gruneisen_strain = epsilon_V * scaling_factor
                        log_messages.extend([
                            f"  - Calculated Bulk Modulus: {B_GPa:.2f} GPa",
                            f"  - Equilibrium Volume: {V0_per_atom * len(symmetrized_atoms):.2f} \u00c5\u00b3",
                            f"  - Base estimated strain: \u00b1{epsilon_V:.4f}",
                            f"  - Exponential scaling factor for strain: {scaling_factor:.3f}",
                            f"  - Final estimated volumetric strain for Gr\u00fcneisen: \u00b1{final_gruneisen_strain:.4f}"
                        ])
                    except Exception as e:
                        final_gruneisen_strain = 0.01
                        log_messages.append(f"  Bulk modulus estimate failed ({e}); fallback to 0.01.")
                else:
                    final_gruneisen_strain = 0.01
                    log_messages.append("  Bulk modulus estimate failed; fallback to 0.01.")
            else:
                log_messages.append(f"User-provided strain for Gr\u00fcneisen: \u00b1{final_gruneisen_strain:.4f}")
            
            for msg in log_messages: print(msg)
            log_path = output_dir / f"gruneisen_strain-{output_prefix}.log"
            log_path.write_text("\n".join(log_messages))

    if not read_fc_path:
        # 1.3 Generate displacements for all phonons
        # 1.3.1 Base phonon
        base_phonon = _pb_generate_displacements(
            input_path=symmetrized_input_path, min_length=min_length, displacement_distance=displacement_distance,
            is_plusminus=is_plusminus, is_diagonal=is_diagonal, tolerance_phonopy=tolerance_phonopy,
            fix_axis=fix_axis, dim_override=dim_override
        )
        
        # 1.3.2 Strained phonons (Optional)
        strained_jobs = []
        if plot_gruneisen:
            original_cell = ase_read(str(symmetrized_input_path), format="vasp")
            gruneisen_dir = output_dir / "gruneisen"
            gruneisen_dir.mkdir(exist_ok=True)
            if selected_axes is None:
                scaling_mode_msg = "Gruneisen cell scaling mode: isotropic (legacy default; all axes scaled uniformly)."
                print(scaling_mode_msg)
            else:
                selected_axes_msg = ",".join(sorted(selected_axes))
                scaling_mode_msg = f"Gruneisen cell scaling mode: selective axes ({selected_axes_msg})."
                print(scaling_mode_msg)
            with open(log_path, "a") as f:
                f.write("\n")
                f.write(scaling_mode_msg + "\n")
                f.write("Lattice constants of strained cells:\n")
            for s in [final_gruneisen_strain, -final_gruneisen_strain]:
                vol_dir = gruneisen_dir / ("plus" if s > 0 else "minus")
                vol_dir.mkdir(exist_ok=True)
                scaled_cell = original_cell.copy()
                if selected_axes is None:
                    # Keep legacy behavior for backward compatibility.
                    scaled_cell.set_cell(original_cell.cell * (1 + s) ** (1 / 3.0), scale_atoms=True)
                else:
                    length_scale = 1.0 + s
                    new_cell = original_cell.cell.copy()
                    scale_a = length_scale if 'a' in selected_axes else 1.0
                    scale_b = length_scale if 'b' in selected_axes else 1.0
                    scale_c = length_scale if 'c' in selected_axes else 1.0
                    new_cell[0] *= scale_a
                    new_cell[1] *= scale_b
                    new_cell[2] *= scale_c
                    scaled_cell.set_cell(new_cell, scale_atoms=True)
                a_len, b_len, c_len = scaled_cell.cell.lengths()
                print(
                    f"  Strain {s:+.4f} -> lattice constants: "
                    f"a={a_len:.6f} A, b={b_len:.6f} A, c={c_len:.6f} A"
                )
                with open(log_path, "a") as f:
                    f.write(
                        f"  Strain {s:+.4f} -> lattice constants: "
                        f"a={a_len:.6f} A, b={b_len:.6f} A, c={c_len:.6f} A\n"
                    )
                strained_poscar_path = vol_dir / "POSCAR"
                ase_write(str(strained_poscar_path), scaled_cell, format="vasp", vasp5=True)
                os.chdir(vol_dir)
                phonon = _pb_generate_displacements(
                    input_path=strained_poscar_path, min_length=min_length, displacement_distance=displacement_distance,
                    is_plusminus=is_plusminus, is_diagonal=is_diagonal, tolerance_phonopy=tolerance_phonopy,
                    fix_axis=fix_axis, dim_override=dim_override
                )
                strained_jobs.append({'strain': s, 'dir': vol_dir, 'phonon': phonon})
                os.chdir(output_dir)

        # 1.4 Batch calculate forces
        all_atoms_for_batch = []
        # Base supercells
        for cell in base_phonon.supercells_with_displacements:
            all_atoms_for_batch.append(Atoms(symbols=cell.symbols, cell=cell.cell, scaled_positions=cell.scaled_positions, pbc=True))
        # Strained supercells
        for job in strained_jobs:
            for cell in job['phonon'].supercells_with_displacements:
                all_atoms_for_batch.append(Atoms(symbols=cell.symbols, cell=cell.cell, scaled_positions=cell.scaled_positions, pbc=True))
                
        all_forces = []
        use_parallel = n_jobs > 1 and macer_device == "cpu"
        if n_jobs > 1 and macer_device != "cpu":
            print(f"WARNING: --n-jobs>1 is supported only with --device cpu (current: {macer_device}). Falling back to serial.")

        if use_parallel:
            tasks = [
                (list(at.symbols), np.array(at.cell), np.array(at.get_scaled_positions()))
                for at in all_atoms_for_batch
            ]
            chunksize = max(1, len(tasks) // (n_jobs * 4))
            with ProcessPoolExecutor(
                max_workers=n_jobs,
                initializer=_init_pb_force_worker,
                initargs=(macer_ff, macer_device, macer_modal, current_macer_model_path),
            ) as pool:
                results_iter = pool.map(_pb_force_worker, tasks, chunksize=chunksize)
                for f in tqdm(results_iter, total=len(tasks), desc="Batch calculating forces"):
                    all_forces.append(f)
        else:
            for atoms in tqdm(all_atoms_for_batch, desc="Batch calculating forces"):
                atoms.calc = workflow_calculator
                all_forces.append(atoms.get_forces())

        # 1.5 Produce Force Constants
        # 1.5.1 Base phonon
        base_num = len(base_phonon.supercells_with_displacements)
        _pb_produce_fc_and_save(base_phonon, all_forces[:base_num])
        
        # 1.5.2 Strained phonons
        offset = base_num
        all_phonons = [base_phonon]
        for job in strained_jobs:
            num = len(job['phonon'].supercells_with_displacements)
            os.chdir(job['dir'])
            _pb_produce_fc_and_save(job['phonon'], all_forces[offset : offset + num])
            all_phonons.append(job['phonon'])
            offset += num
            os.chdir(output_dir)

    # Apply mass override if provided
    if mass_map:
        print("\n--- Applying mass override ---")
        for ph in all_phonons:
            symbols = ph.primitive.symbols
            ph.masses = [mass_map.get(s, m) for s, m in zip(symbols, ph.masses)]
        print(f"  Applied mass override: {mass_map}")

    print("\n--- Step 4: Creating band.conf ---")
    band_conf_path = output_dir / (out_path_arg or "band.conf")
    disp_yaml_path = output_dir / (yaml_path_arg or "phonopy_disp.yaml")
    try:
        generate_band_conf(
            poscar_path=symmetrized_input_path, yaml_path=disp_yaml_path, out_path=band_conf_path,
            gamma_label=gamma_label, symprec=symprec_seekpath, dim_override=dim_override,
            no_defaults=no_defaults_band_conf, atom_names_override=atom_names_override,
            rename_override=rename_override,
            manual_band_path=manual_band_path,
            manual_labels=manual_labels,
        )
        print(f"{band_conf_path.name} created successfully.")
    except Exception:
        import traceback
        print("Error during band.conf generation:", traceback.format_exc())
        return

    # === Step 5: Plotting band structure ===
    print("\n--- Step 5: Plotting band structure ---")
    band_pdf_path = output_dir / "band.pdf"
    band_yaml_path = output_dir / "band.yaml"
    
    try:
        print(f"Generating band structure from {band_conf_path.name} using phonopy API...")
        
        # 1. Read band.conf settings using PhonopyConfParser
        from phonopy.cui.settings import PhonopyConfParser
        conf_parser = PhonopyConfParser(filename=str(band_conf_path))
        settings = conf_parser.settings

        # 2. Get q-points for the band path
        npoints = settings.band_points
        if npoints is None:
            npoints = 51  # Default value in phonopy
        
        qpoints = get_band_qpoints(
            settings.band_paths,
            npoints=npoints
        )
        labels = settings.band_labels
        path_connections = []
        for paths in settings.band_paths:
            path_connections += [True, ] * (len(paths) - 2)
            path_connections.append(False)

        # 3. Run calculation on the existing phonon object
        base_phonon.run_band_structure(qpoints,
                                       path_connections=path_connections,
                                       labels=labels,
                                       with_eigenvectors=with_eigenvectors)

        # 4. Save YAML and Plot
        base_phonon.write_yaml_band_structure(filename=str(band_yaml_path))
        print(f"{band_yaml_path.name} created successfully.")
        
        band_plot = base_phonon.plot_band_structure()
        band_plot.savefig(str(band_pdf_path))
        print(f"{band_pdf_path.name} created successfully.")

        # 5. Write .dat file (existing logic)
        band_dat_path = output_dir / "band.dat"
        _write_band_dat(band_yaml_path, band_dat_path)

    except Exception as e:
        print("Error during API-based band structure plotting:", e)
        import traceback
        traceback.print_exc()
        return

    # === Step 5.1: Phonon DOS (optional) ===
    if plot_dos:
        print("\n--- Step 5.1: Calculating and Plotting Phonon DOS ---")
        try:
            # We use default filenames initially, then rename them in Step 6
            # New unified filenames
            dos_dat_filename = "phonon_dos.dat"
            dos_pdf_filename = "phonon_dos.pdf"
            
            # For PDOS, we need eigenvectors and no mesh symmetry
            print(f"Calculating DOS with mesh: {mesh} (with_eigenvectors=True, is_mesh_symmetry=False)")
            base_phonon.run_mesh(mesh, with_eigenvectors=True, is_mesh_symmetry=False)
            
            # 1. Calculate Total DOS
            # Run without args first to let it auto-determine range
            base_phonon.run_total_dos()
            total_dos_dict = base_phonon.get_total_dos_dict()
            freqs = total_dos_dict['frequency_points']
            total_dos_val = total_dos_dict['total_dos']
            
            # Extract frequency grid parameters to ensure PDOS matches exactly
            f_min = freqs[0]
            f_max = freqs[-1]
            n_points = len(freqs)
            print(f"DOS frequency range: {f_min:.4f} to {f_max:.4f} (n_points={n_points})")

            # 2. Calculate Projected DOS (PDOS) and Group by Element
            # run_projected_dos does NOT take mesh parameters in API, it uses the mesh object's settings
            base_phonon.run_projected_dos()
            pdos_dict = base_phonon.get_projected_dos_dict()
            raw_pdos = np.array(pdos_dict['projected_dos'])
            
            print(f"DEBUG: raw_pdos shape: {raw_pdos.shape}")
            # Ensure shape is (n_freqs, n_atoms)
            # If shape is (n_atoms, n_freqs) [which is (16, 201) in this case], transpose it
            if raw_pdos.shape[0] == len(base_phonon.primitive) and raw_pdos.shape[1] != len(base_phonon.primitive):
                 print("DEBUG: Transposing raw_pdos to (n_freqs, n_atoms)")
                 raw_pdos = raw_pdos.T
            
            primitive = base_phonon.primitive
            symbols = primitive.symbols
            unique_symbols = sorted(list(set(symbols)))
            print(f"Grouping PDOS by elements: {unique_symbols}")
            
            grouped_pdos = {}
            for sym in unique_symbols:
                indices = [i for i, s in enumerate(symbols) if s == sym]
                # Sum over atoms of the same element
                # raw_pdos has shape (n_freq, n_atoms)
                grouped_pdos[sym] = np.sum(raw_pdos[:, indices], axis=1)

            # Ensure PDOS frequency grid matches Total DOS
            # If they differ slightly in length (e.g. Total DOS has auto-grid), we might need to interpolation or just use PDOS grid
            # But usually run_total_dos and run_projected_dos on same mesh should be consistent.
            # Let's check lengths.
            if len(freqs) != len(grouped_pdos[unique_symbols[0]]):
                print(f"WARNING: Frequency point mismatch! Total: {len(freqs)}, PDOS: {len(grouped_pdos[unique_symbols[0]])}")
                # Fallback: Use PDOS frequency points if available
                if 'frequency_points' in pdos_dict:
                     freqs = pdos_dict['frequency_points']
                     # Re-calculate Total DOS on this grid? 
                     # Or just assume Total DOS is close enough? 
                     # If lengths differ, we can't write them in same file side-by-side easily without processing.
                     # But typically they match.
            
            # 3. Write Combined Data File
            header = f"# Tetrahedron method\n# Frequency          Total               {'              '.join(unique_symbols)}\n"
            
            with open(dos_dat_filename, 'w') as f:
                f.write(header)
                for i, freq in enumerate(freqs):
                    line_parts = [f"{freq:18.10f}", f"{total_dos_val[i]:18.10f}"]
                    for sym in unique_symbols:
                        line_parts.append(f"{grouped_pdos[sym][i]:18.10f}")
                    f.write("".join(line_parts) + "\n")
            print(f"{dos_dat_filename} created successfully.")
            
            # 4. Plot Combined DOS (Total + Projected)
            fig, ax = plt.subplots()
            ax.plot(freqs, total_dos_val, 'k-', linewidth=1.5, label='Total')
            
            # Use a colormap or cycle for different elements
            colors = plt.cm.tab10(np.linspace(0, 1, len(unique_symbols)))
            for i, sym in enumerate(unique_symbols):
                ax.plot(freqs, grouped_pdos[sym], linewidth=1.0, label=sym, color=colors[i])
            
            ax.set_xlabel('Frequency (THz)')
            ax.set_ylabel('Density of States')
            ax.legend(loc='best')
            ax.grid(True, linestyle='--', alpha=0.6)
            
            plt.tight_layout()
            plt.savefig(dos_pdf_filename)
            plt.close(fig)
            print(f"{dos_pdf_filename} created successfully (Total + Projected).")
            
        except Exception as e:
            print("Error during DOS calculation/plotting:", e)
            import traceback
            traceback.print_exc()

    # === Step 5.5: Irreducible Representations (optional) ===
    if show_irreps:
        print("\n--- Step 5.5: Calculating Irreducible Representations ---")
        try:
            q_point = irreps_qpoint if irreps_qpoint else [0.0, 0.0, 0.0]
            
            # Determine tolerance for irreps
            # If tolerance_irreps is not provided, use tolerance_phonopy
            irreps_tol = tolerance_irreps if tolerance_irreps is not None else tolerance_phonopy
            
            print(f"Calculating irreps at q-point: {q_point} (tolerance: {irreps_tol})")
            
            # Using set_irreps and show_irreps as verified in Phonopy API
            base_phonon.set_irreps(
                q_point,
                is_little_cogroup=False,
                nac_q_direction=None,
                degeneracy_tolerance=irreps_tol
            )
            base_phonon.show_irreps() # Prints to stdout
            
            # Write to yaml file
            base_phonon.write_yaml_irreps(show_irreps=True)

        except Exception as e:
            print("Error during irreducible representation calculation:", e)
            import traceback
            traceback.print_exc()

    # === Step 5.6: Gruneisen workflow (optional) ===
    if plot_gruneisen:
        # Construct meta_path
        meta_path = "N/A"
        if labels:
            flat_labels = []
            for item in labels:
                if isinstance(item, list): flat_labels.extend(item)
                else: flat_labels.append(item)
            
            # Simple dedup for adjacent same labels (e.g. G-X, X-M -> G-X-M)
            clean_path = []
            for l in flat_labels:
                l_clean = l.replace("GAMMA", "Γ").replace("GM", "Γ")
                if not clean_path or clean_path[-1] != l_clean:
                    clean_path.append(l_clean)
            meta_path = "-".join(clean_path)

        dat_filename, plot_distances, plot_segment_nqpoint, plot_labels = _run_gruneisen_on_top(
            phonons=all_phonons,
            strain=final_gruneisen_strain,
            band_yaml_path=band_yaml_path,
            output_prefix=output_prefix,
            mesh=mesh,
            meta_formula=initial_atoms.get_chemical_formula(),
            meta_sg=final_symmetrized_sg_symbol,
            meta_path=meta_path
        )
        if dat_filename and dat_filename.exists():
            pdf_filename = f"gruneisen_band-{output_prefix}.pdf"
            
            # Capture consistent limits from the main band plot
            gmin_used, gmax_used = _plot_gruneisen_from_dat(
                dat_filename=dat_filename, 
                distances=plot_distances,
                segment_nqpoint=plot_segment_nqpoint,
                labels=plot_labels,
                gmin=gmin, gmax=gmax,
                output_filename=pdf_filename,
                filter_outliers_factor=filter_outliers_factor,
            )
            
            mode_pdf_filename = f"gruneisen_mode_parameter-{output_prefix}.pdf"
            _plot_mode_gruneisen_from_dat(
                dat_filename=dat_filename, 
                distances=plot_distances,
                segment_nqpoint=plot_segment_nqpoint,
                labels=plot_labels,
                gmin=gmin_used, gmax=gmax_used,
                output_filename=mode_pdf_filename,
                filter_outliers_factor=filter_outliers_factor,
            )
            
            # 3D Visualization (Full Mesh)
            from macer.utils.viz_tools import plot_gruneisen_3d_style
            full_dat_path = output_dir / f"gruneisen_full-{output_prefix}.dat"
            if full_dat_path.exists():
                print("Generating 3D Grüneisen visualization...")
                for suffix, mode in [("average", "mean"), ("minima", "min")]:
                    plot_gruneisen_3d_style(
                        full_dat_path,
                        poscar_path=symmetrized_input_path,
                        out_prefix=f"gruneisen_full-{output_prefix}_{suffix}",
                        cmin=gmin if gmin is not None else -5.0,
                        cmax=gmax if gmax is not None else 5.0,
                        cut_f=0.0,
                        original_position=False,
                        fill_bz=False,
                        plot_rc=False,
                        expand_symmetry=True,
                        full_bz=False,
                        special_points=True,
                        mode=mode,
                    )

    # === Step 6: Renaming outputs ===
    print("\n--- Step 6: Renaming outputs ---")
    
    # Files to simply rename
    rename_map = {
        "phonopy_disp.yaml": f"phonopy_disp-{output_prefix}.yaml",
        band_conf_path.name: f"band-{output_prefix}.conf",
        band_pdf_path.name: f"band-{output_prefix}.pdf",
        band_yaml_path.name: f"band-{output_prefix}.yaml",
        "band.dat": f"band-{output_prefix}.dat",
        "irreps.yaml": f"irreps-{output_prefix}.yaml",
        "phonon_dos.dat": f"phonon_dos-{output_prefix}.dat",
        "phonon_dos.pdf": f"phonon_dos-{output_prefix}.pdf",
    }

    for old, new in rename_map.items():
        old_path = output_dir / old
        if old_path.exists():
            try:
                if old == "phonopy_disp.yaml":
                    shutil.copy(old_path, output_dir / new)
                    print(f"Copied {old} -> {new}")
                else:
                    old_path.rename(output_dir / new)
                    print(f"Renamed {old} -> {new}")
            except FileNotFoundError:
                print(f"Warning: Could not find {old} to rename.")
    
    # Update paths to point to renamed files for subsequent steps
    band_conf_path = output_dir / f"band-{output_prefix}.conf"
    band_yaml_path = output_dir / f"band-{output_prefix}.yaml"

    # Files to copy and keep original (FORCE_SETS, FORCE_CONSTANTS, SPOSCAR)
    copy_map = {
        "FORCE_SETS": f"FORCE_SETS_{output_prefix}",
        "FORCE_CONSTANTS": f"FORCE_CONSTANTS_{output_prefix}",
    }
    
    # Check for SPOSCAR and add to copy map
    if (output_dir / "SPOSCAR").exists():
        copy_map["SPOSCAR"] = f"SPOSCAR_{output_prefix}"

    for original, copy_name in copy_map.items():
        original_path = output_dir / original
        if original_path.exists():
            try:
                shutil.copy(original_path, output_dir / copy_name)
                print(f"Copied {original} -> {copy_name}")
            except Exception as e:
                print(f"Warning: Could not copy {original}: {e}")

    # Copy input POSCAR to POSCAR-input and POSCAR
    if input_path.exists():
        try:
            shutil.copy(input_path, output_dir / "POSCAR-input")
            shutil.copy(input_path, output_dir / "POSCAR")
            print(f"Copied input {input_path.name} -> POSCAR-input and POSCAR")
        except Exception as e:
            print(f"Warning: Could not copy input POSCAR: {e}")

    # === Step 7: Write VESTA files for arrows (optional) ===
    if write_arrow or show_irreps:
        write_arrow = True # Force True if show_irreps is on
        print("\n--- Step 7: Writing VESTA files for phonon modes ---")
        try:
            from macer.phonopy.write_arrow_phonopy import write_vesta_files_for_arrows
            
            # 1. Determine unique output directory
            base_arrow_dir = output_dir / f"ARROW-{output_prefix}"
            arrow_dir = base_arrow_dir
            if arrow_dir.exists():
                i = 1
                while arrow_dir.exists():
                    arrow_dir = output_dir / f"ARROW-{output_prefix}-NEW{i:02d}"
                    i += 1
            arrow_dir.mkdir(parents=True)

            # 2. Determine execution targets
            target_points = []
            
            # Pinpoint mode: Calculate specific points
            if show_irreps:
                q_irr = np.array(irreps_qpoint) if irreps_qpoint else np.array([0.0, 0.0, 0.0])
                target_points.append((q_irr, "IRR", 1))
                print(f"  Mode: Writing arrows for Irreps Q-point: {q_irr}")

            elif arrow_qpoint_gamma:
                 target_points.append((np.array([0.0, 0.0, 0.0]), "GM", 1))
                 print("  Mode: Writing arrows for Gamma point only.")
            
            elif arrow_qpoint is not None:
                 q_vec = np.array(arrow_qpoint)
                 target_points.append((q_vec, "SELECTIVE", 1))
                 print(f"  Mode: Writing arrows for user-specified Q-point: {q_vec}")
            
            else:
                 # Default: Special Points from band.conf
                 print("  Mode: Writing arrows for Special Q-points detected in band path.")
                 from phonopy.cui.settings import PhonopyConfParser
                 conf_parser = PhonopyConfParser(filename=str(band_conf_path))
                 settings = conf_parser.settings
                 
                 q_sequence = []
                 l_sequence = []
                 
                 def clean_l(s):
                     s = s.strip()
                     # Handle LaTeX and common Gamma variations
                     if s.upper() in ["GAMMA", "G", "GM", r"$\Gamma$", r"$\GAMMA$", r"\GAMMA"]:
                         return "GM"
                     return s.replace("$", "").replace("\\", "").replace("_", "")

                 if settings.band_paths:
                     labels_flat = []
                     if settings.band_labels:
                         for l_seg in settings.band_labels:
                             if isinstance(l_seg, list): labels_flat.extend(l_seg)
                             else: labels_flat.append(l_seg)
                     
                     q_idx_in_flat = 0
                     for path_seg in settings.band_paths:
                         for q in path_seg:
                             q_np = np.array(q)
                             l_raw = labels_flat[q_idx_in_flat] if q_idx_in_flat < len(labels_flat) else "QP"
                             l_c = clean_l(l_raw)
                             
                             is_duplicate = False
                             for sq, sl in zip(q_sequence, l_sequence):
                                 if np.allclose(q_np, sq, atol=1e-4) and l_c == sl:
                                     is_duplicate = True
                                     break
                             
                             if not is_duplicate:
                                 q_sequence.append(q_np)
                                 l_sequence.append(l_c)
                             q_idx_in_flat += 1
                 
                 for i, (q, l) in enumerate(zip(q_sequence, l_sequence)):
                     target_points.append((q, l, i+1))

            # Execute Loop for Target Points
            for q_pos, label, idx in target_points:
                q_str = f"{q_pos[0]:.3f}_{q_pos[1]:.3f}_{q_pos[2]:.3f}"
                
                # If Irreps mode and only one point, write directly to arrow_dir
                if show_irreps and len(target_points) == 1:
                    sub_dir = arrow_dir
                    print(f"  - Processing Q-point: {q_pos} (writing directly to {arrow_dir.name})")
                elif label == "SELECTIVE":
                    sub_dir_name = f"QPOINTS_SELECTIVE={q_str}"
                    sub_dir = arrow_dir / sub_dir_name
                    sub_dir.mkdir(parents=True, exist_ok=True)
                    print(f"  - Processing Q-point: {label} {q_pos}")
                else:
                    sub_dir_name = f"QPOINTS_{idx:03d}-{label}={q_str}"
                    sub_dir = arrow_dir / sub_dir_name
                    sub_dir.mkdir(parents=True, exist_ok=True)
                    print(f"  - Processing Q-point {idx}: {label} {q_pos}")
                
                # Run q-point calculation
                base_phonon.run_qpoints([q_pos], with_eigenvectors=True)
                
                # Write temp qpoints.yaml
                base_phonon.write_yaml_qpoints_phonon() # Writes to qpoints.yaml
                
                temp_yaml = output_dir / "qpoints.yaml"
                if not temp_yaml.exists():
                     print("    Error: qpoints.yaml not generated.")
                     continue
                
                # If this is a pinpoint case with a single point, save as eigenvector-{prefix}.yaml for convenience
                if (arrow_qpoint_gamma or show_irreps or arrow_qpoint is not None) and len(target_points) == 1:
                    eigenvector_yaml_path = output_dir / f"eigenvector-{output_prefix}.yaml"
                    shutil.copy(temp_yaml, eigenvector_yaml_path)
                    print(f"    Saved eigenvector data to {eigenvector_yaml_path.name}")

                # Generate VESTA
                irreps_yaml_path_to_pass = output_dir / f"irreps-{output_prefix}.yaml" if show_irreps else None
                write_vesta_files_for_arrows(
                    band_yaml_path=temp_yaml,
                    poscar_path=symmetrized_input_path, # Use the actual cell used for calculation
                    output_dir=sub_dir,
                    arrow_length=arrow_length,
                    arrow_min_cutoff=arrow_min_cutoff,
                    target_q_point=q_pos, 
                    irreps_yaml_path=irreps_yaml_path_to_pass
                )
                
                if temp_yaml.exists():
                    temp_yaml.unlink() # Cleanup temp file

        except Exception as e:
            print("Error during VESTA file generation for arrows:", e)
            import traceback
            traceback.print_exc()

    # === Step 8: Write modulation-example.conf ===
    print("\n--- Step 8: Writing modulation-example.conf ---")
    dim_str_mod = " ".join(map(str, base_phonon.supercell_matrix.diagonal().astype(int)))
    if np.count_nonzero(base_phonon.supercell_matrix - np.diag(base_phonon.supercell_matrix.diagonal())) != 0:
         dim_str_mod = " ".join(map(str, base_phonon.supercell_matrix.flatten().astype(int)))

    modulation_content = f"""# supercell dim
# qpoint
# band index
# amplitude
# degree
DIM = {dim_str_mod}
MODULATION = 2 2 1, 0.5 0.5 0.0 1 10 0
#MODULATION = 3 3 1, 0.0 0.0 0.0 6 5 0
#FORCE_CONSTANTS = READ
"""
    (output_dir / "modulation-example.conf").write_text(modulation_content)
    print(f"Created modulation-example.conf")

    mesh_gamma_raman_content = f"""# mesh-gamma-phonopy-raman.conf
DIM = {dim_str_mod}
MP = 1 1 1

# Read "FORCE_CONSTANTS"
FORCE_CONSTANTS = READ
# If read "FORCE_SETS"
# FORCE_SETS = READ

# For phonon displacement
EIGENVECTORS = .TRUE.

# output file format
MESH_FORMAT = YAML
"""
    (output_dir / "mesh-gamma-phonopy-raman.conf").write_text(mesh_gamma_raman_content)
    print("Created mesh-gamma-phonopy-raman.conf")

    # === Cleanup ===
    print("\n--- Cleaning up ---")
    cleanup_patterns = ['vasprun-*.xml', 'OUTCAR-*', 'CONTCAR-*', 'relax-*.log.*', 'phonopy.yaml']
    for pattern in cleanup_patterns:
        for item in output_dir.glob(pattern):
            try:
                if item.is_dir(): shutil.rmtree(item)
                else: item.unlink()
            except OSError as e:
                print(f"Error cleaning up {item}: {e}")

    print(f"\nWorkflow for {input_path.name} completed.")
    print("\n--- Final summary ---")
    print(f"  Space group of {symmetrized_input_path.name}: {final_symmetrized_sg_symbol} (No. {final_symmetrized_sg_number})")
    
    # Get the supercell dimension string for the help message
    dim_str_help = " ".join(map(str, base_phonon.supercell_matrix.diagonal().astype(int)))
    if np.count_nonzero(base_phonon.supercell_matrix - np.diag(base_phonon.supercell_matrix.diagonal())) != 0:
         dim_str_help = " ".join(map(str, base_phonon.supercell_matrix.flatten().astype(int)))

    print("\n--- Post-processing Help ---")
    print(f"Calculation directory: {output_dir}")
    print("To re-calculate irreducible representations (example):")
    print(f'  phonopy --dim="{dim_str_help}" --irreps="0 0 0" -c POSCAR-input --tolerance {tolerance_phonopy}')
    print("To re-plot band structure (example):")
    print(f'  phonopy --dim="{dim_str_help}" -c POSCAR-input -p -s band-{output_prefix}.conf')

    print("===== macer phonon workflow: DONE =====\n")



"""O2 reps of space group ops with respect to atomic coordinate basis."""

from __future__ import annotations

from typing import Optional

import numpy as np
from scipy.sparse import csr_array

from macer.externals.symfc_bundled.utils.utils import SymfcAtoms

from .spg_reps_base import SpgRepsBase


class SpgRepsO2(SpgRepsBase):
    """Class of reps of space group operations for fc2."""

    def __init__(
        self, supercell: SymfcAtoms, spacegroup_operations: Optional[dict] = None
    ):
        """Init method.

        Parameters
        ----------
        supercell : SymfcAtoms
            Supercell.
        spacegroup_operations : dict, optional
            Space group operations in supercell, by default None. When None,
            spglib is used. The following keys and values correspond to spglib
            symmetry dataset:
                rotations : array_like
                translations : array_like

        """
        self._r2_reps: list[csr_array]
        super().__init__(supercell, spacegroup_operations=spacegroup_operations)

    @property
    def r_reps(self) -> list[csr_array]:
        """Return 2nd rank tensor rotation matricies."""
        return self._r2_reps

    def get_sigma2_rep(self, i: int, nonzero: Optional[np.ndarray] = None) -> csr_array:
        """Compute vector representation of i-th atomic pair permutation matrix.

        Parameters
        ----------
        i : int
            Index of coset presentations of space group operations.

        """
        return self._get_sigma2_rep_data(i, nonzero=nonzero)

    def _prepare(self, spacegroup_operations):
        super()._prepare(spacegroup_operations)
        N = len(self._numbers)
        self._atom_pairs = (np.mgrid[0:N, 0:N].reshape((2, -1)).T).astype(
            "uint16", copy=False
        )
        self._coeff = np.array([N, 1], dtype=int)
        self._compute_r2_reps()

    def _compute_r2_reps(self, tol: float = 1e-10):
        """Compute and return 2nd rank tensor rotation matricies."""
        r2_reps = []
        for r in self._unique_rotations:
            r_c = self._lattice.T @ r @ np.linalg.inv(self._lattice.T)
            r2_rep = np.kron(r_c, r_c)
            row, col = np.nonzero(np.abs(r2_rep) > tol)
            data = r2_rep[(row, col)]
            r2_reps.append(csr_array((data, (row, col)), shape=r2_rep.shape))
        self._r2_reps = r2_reps

    def _get_sigma2_rep_data(
        self, i: int, nonzero: Optional[np.ndarray] = None
    ) -> csr_array:
        """Compute vector representation of i-th atomic pair permutation matrix.

        Operation permutation[self._atom_pairs @ self._coeff is divided
        to reduce memory allocation.

        permutation_pairs represents row indices of nonzero elements
        in permutation matrix. Column indices of nonzero elements in permutation
        matrix are array indices of permutation_pairs.
        """
        uri = self._unique_rotation_indices
        permutation = self._permutations[uri[i]]
        if nonzero is not None:
            pairs = self._atom_pairs[nonzero]
        else:
            pairs = self._atom_pairs
        permutation_pairs = permutation[pairs[:, 0]] * self._coeff[0]
        permutation_pairs += permutation[pairs[:, 1]]
        return permutation_pairs


class SpgRepsO2MatrixReps(SpgRepsBase):
    """Class of reps of space group operations for fc2."""

    def __init__(
        self, supercell: SymfcAtoms, spacegroup_operations: Optional[dict] = None
    ):
        """Init method.

        Parameters
        ----------
        supercell : SymfcAtoms
            Supercell.
        spacegroup_operations : dict, optional
            Space group operations in supercell, by default None. When None,
            spglib is used. The following keys and values correspond to spglib
            symmetry dataset:
                rotations : array_like
                translations : array_like

        """
        self._r2_reps: list[csr_array]
        self._col: np.ndarray
        self._data: np.ndarray
        super().__init__(supercell, spacegroup_operations=spacegroup_operations)

    @property
    def r_reps(self) -> list[csr_array]:
        """Return 2nd rank tensor rotation matricies."""
        return self._r2_reps

    def get_sigma2_rep(self, i: int) -> csr_array:
        """Compute and return i-th atomic pair permutation matrix.

        Parameters
        ----------
        i : int
            Index of coset presentations of space group operations.

        """
        data, row, col, shape = self._get_sigma2_rep_data(i)
        return csr_array((data, (row, col)), shape=shape)

    def _prepare(self, spacegroup_operations):
        super()._prepare(spacegroup_operations)
        N = len(self._numbers)
        a = np.arange(N)
        self._atom_pairs = np.stack(np.meshgrid(a, a), axis=-1).reshape(-1, 2)
        self._coeff = np.array([1, N], dtype=int)
        self._col = self._atom_pairs @ self._coeff
        self._data = np.ones(N * N, dtype=int)
        self._compute_r2_reps()

    def _compute_r2_reps(self, tol: float = 1e-10):
        """Compute and return 2nd rank tensor rotation matricies."""
        r2_reps = []
        for r in self._unique_rotations:
            r_c = self._lattice.T @ r @ np.linalg.inv(self._lattice.T)
            r2_rep = np.kron(r_c, r_c)
            row, col = np.nonzero(np.abs(r2_rep) > tol)
            data = r2_rep[(row, col)]
            r2_reps.append(csr_array((data, (row, col)), shape=r2_rep.shape))
        self._r2_reps = r2_reps

    def _get_sigma2_rep_data(self, i: int) -> tuple:
        uri = self._unique_rotation_indices
        permutation = self._permutations[uri[i]]
        NN = len(self._numbers) ** 2
        row = permutation[self._atom_pairs] @ self._coeff
        return self._data, row, self._col, (NN, NN)



"""Symmetry adapted basis sets of force constants."""

from __future__ import annotations

from abc import ABC, abstractmethod

from numpy.typing import NDArray

from macer.externals.symfc_bundled.spg_reps import SpgRepsBase
from macer.externals.symfc_bundled.utils.cutoff_tools import FCCutoff
from macer.externals.symfc_bundled.utils.matrix import BlockMatrixNode
from macer.externals.symfc_bundled.utils.utils import SymfcAtoms


class FCBasisSetBase(ABC):
    """Abstract base class of symmetry adapted basis set for force constants."""

    def __init__(
        self,
        supercell: SymfcAtoms,
        cutoff: float | None = None,
        use_mkl: bool = False,
        log_level: int = 0,
    ):
        """Init method.

        Parameters
        ----------
        supercell : SymfcAtoms
            Supercell.
        cutoff: float
            Cutoff distance in angstroms. Default is None.
        use_mkl : bool
            Use MKL or not. Default is False.
        log_level : int, optional
            Log level. Default is 0.

        """
        self._supercell = supercell
        self._natom = len(supercell)
        self._use_mkl = use_mkl
        self._log_level = log_level
        self._spg_reps: SpgRepsBase
        self._atomic_decompr_idx: NDArray
        self._basis_set: NDArray | None
        self._blocked_basis_set: BlockMatrixNode | None

        if cutoff is None:
            self._fc_cutoff = None
        else:
            self._fc_cutoff = FCCutoff(supercell, cutoff=cutoff)

    @property
    @abstractmethod
    def compact_compression_matrix(self) -> NDArray | None:
        """Return compression matrix for compact basis set."""
        pass

    @property
    @abstractmethod
    def compression_matrix(self) -> NDArray | None:
        """Return compression matrix."""
        pass

    @property
    def basis_set(self) -> NDArray:
        """Return compressed basis set.

        shape=(n_c, n_bases), dtype='double'.

        """
        if self._blocked_basis_set is None:
            raise ValueError("Basis set is not computed yet.")
        return self._blocked_basis_set.recover()

    @property
    def blocked_basis_set(self) -> BlockMatrixNode | None:
        """Return compressed basis set in blocked format."""
        return self._blocked_basis_set

    @property
    def atomic_decompr_idx(self) -> NDArray:
        """Return atomic permutations by lattice translations."""
        return self._atomic_decompr_idx

    @property
    def translation_permutations(self) -> NDArray:
        """Return permutations by lattice translation."""
        if self._spg_reps is None:
            raise ValueError("SpgRepsBase is not set.")
        return self._spg_reps.translation_permutations

    @property
    def p2s_map(self) -> NDArray:
        """Return indices of translationally independent atoms."""
        if self._spg_reps is None:
            raise ValueError("SpgRepsBase is not set.")
        if self._spg_reps.p2s_map is None:
            raise ValueError("p2s_map is not set.")
        return self._spg_reps.p2s_map

    @property
    def fc_cutoff(self) -> FCCutoff | None:
        """Return force constants cutoff."""
        return self._fc_cutoff

    @abstractmethod
    def run(self):
        """Run basis set calculation."""
        pass



"""Permutation utility functions."""

from typing import Optional

import numpy as np
import scipy
from scipy.sparse import csr_array

from macer.externals.symfc_bundled.utils.cutoff_tools import FCCutoff


def get_entire_combinations(n: int, r: int):
    """Return numpy array of combinations.

    combinations = np.array(
       list(itertools.combinations(range(n), r)), dtype=int
    )
    """
    combs = np.ones((r, n - r + 1), dtype=int)
    combs[0] = np.arange(n - r + 1)
    for j in range(1, r):
        reps = (n - r + j) - combs[j - 1]
        combs = np.repeat(combs, reps, axis=1)
        ind = np.add.accumulate(reps)
        combs[j, ind[:-1]] = 1 - reps[1:]
        combs[j, 0] = j
        combs[j] = np.add.accumulate(combs[j])
    return combs.T


def get_combinations(
    natom: int,
    order: int,
    fc_cutoff: Optional[FCCutoff] = None,
    indep_atoms: Optional[np.ndarray] = None,
):
    """Return numpy array of FC index combinations."""
    if fc_cutoff is None:
        combinations = get_entire_combinations(3 * natom, order)
    else:
        if order == 2:
            combinations = fc_cutoff.combinations2()
        elif order == 3:
            """Combinations can be divided using fc_cut.combiations3(i)."""
            combinations = fc_cutoff.combinations3_all()
        elif order == 4:
            combinations = fc_cutoff.combinations4_all()
        else:
            raise NotImplementedError(
                "Combinations are implemented only for 2 <= order <= 4."
            )

    if indep_atoms is not None:
        nonzero = np.zeros(combinations.shape[0], dtype=bool)
        atom_indices = combinations[:, 0] // 3
        for i in indep_atoms:
            nonzero[atom_indices == i] = True
        combinations = combinations[nonzero]
    return combinations


def _eliminate_zero_elements(
    perm_decompr_idx: np.ndarray, nonzero: np.ndarray
) -> np.ndarray:
    """Eliminate zero elements and reindex orbit indexes."""
    size_full = len(perm_decompr_idx)
    if not np.all(nonzero):
        perm_decompr_idx = perm_decompr_idx[nonzero]
        nonzero_map = np.ones(size_full, dtype="int") * -1
        nonzero_map[nonzero] = np.arange(len(perm_decompr_idx))
        perm_decompr_idx = nonzero_map[perm_decompr_idx]
    return perm_decompr_idx


def construct_basis_from_perm_decompr_indices(
    perm_decompr_idx: np.ndarray, verbose: bool = False
):
    """Transform perm_decompr_idx into basis matrix.

    Parameters
    ----------
    perm_decompr_idx: Decompression indices of lattice translation basis
                      using permutations.
    Return
    ------
    c_pt: Compressed basis matrix for permutations and lattice translations.
          c_pt = eigh(C_trans.T @ C_perm @ C_perm.T @ C_trans)
    """
    if verbose:
        print("Construct permutation basis matrix.", flush=True)

    size_full = len(perm_decompr_idx)
    nonzero = perm_decompr_idx != -1
    perm_decompr_idx = _eliminate_zero_elements(perm_decompr_idx, nonzero)

    size1 = len(perm_decompr_idx)
    perm_lat_trans_graph = csr_array(
        (np.ones(size1, dtype=bool), (np.arange(size1), perm_decompr_idx)),
        shape=(size1, size1),
        dtype=bool,
    )

    n_col, cols = scipy.sparse.csgraph.connected_components(perm_lat_trans_graph)
    key, cnt = np.unique(cols, return_counts=True)
    values = np.reciprocal(np.sqrt(cnt))

    rows = np.where(nonzero)[0]
    c_pt = csr_array(
        (values[cols], (rows, cols)),
        shape=(size_full, n_col),
        dtype="double",
    )
    return c_pt



import os
import numpy as np
import mmap
import macer.externals.dynaphopy_bundled.dynamics as dyn
import warnings


# VASP OUTCAR file parser
def read_vasp_trajectory(file_name, structure=None, time_step=None,
                         limit_number_steps=10000000,  # Maximum number of steps read (for security)
                         last_steps=None,
                         initial_cut=1,
                         end_cut=None,
                         memmap=False,
                         template=None):

    # warning
    warnings.warn('This parser will be deprecated, you can use XDATCAR instead', DeprecationWarning)

    # Check file exists
    if not os.path.isfile(file_name):
        print('Trajectory file does not exist!')
        exit()

    # Check time step
    if time_step is not None:
        print('Warning! Time step flag has no effect reading from VASP OUTCAR file (time step will be read from file)')

    if memmap:
        print('Warning! Memory mapping is not implemented in VASP OUTCAR parser')

    # Starting reading
    print("Reading VASP trajectory")
    print("This may take long, please wait..")

    # Dimensionality of VASP calculation
    number_of_dimensions = 3

    with open(file_name, "r+") as f:

        #Memory-map the file
        file_map = mmap.mmap(f.fileno(), 0)
        position_number=file_map.find(b'NIONS =')
        file_map.seek(position_number+7)
        number_of_atoms = int(file_map.readline())

        #Read time step
        position_number=file_map.find(b'POTIM  =')
        file_map.seek(position_number+8)
        time_step = float(file_map.readline().split()[0])* 1E-3 # in picoseconds

        #Reading super cell
        position_number = file_map.find(b'direct lattice vectors')
        file_map.seek(position_number)
        file_map.readline()
        super_cell = []
        for i in range (number_of_dimensions):
            super_cell.append(file_map.readline().split()[0:number_of_dimensions])
        super_cell = np.array(super_cell, dtype='double')

        file_map.seek(position_number)
        file_map.readline()

    # Check if number of atoms is multiple of cell atoms
        if structure is not None:
            if number_of_atoms % structure.get_number_of_cell_atoms() != 0:
                print('Warning: Number of atoms not matching, check VASP output files')
    #        structure.set_number_of_atoms(number_of_atoms)

#       Read coordinates and energy
        trajectory = []
        energy = []
        counter = 0
        while True:

            counter +=1
            #Initial cut control
            if initial_cut > counter:
                continue

            position_number=file_map.find(b'POSITION')
            if position_number < 0 : break

            file_map.seek(position_number)
            file_map.readline()
            file_map.readline()

            read_coordinates = []
            for i in range (number_of_atoms):
                read_coordinates.append(file_map.readline().split()[0:number_of_dimensions])

            read_coordinates = np.array(read_coordinates, dtype=float) # in angstrom
            if template is not None:
                indexing = np.argsort(template)
                read_coordinates = read_coordinates[indexing, :]

            position_number=file_map.find(b'energy(')
            file_map.seek(position_number)
            read_energy = file_map.readline().split()[2]
            trajectory.append(read_coordinates.flatten()) #in angstrom
            energy.append(np.array(read_energy, dtype=float))

            #security routine to limit maximum of steps to read and put in memory
            if limit_number_steps+initial_cut < counter:
                print("Warning! maximum number of steps reached! No more steps will be read")
                break

            if end_cut is not None and end_cut <= counter:
                break

        file_map.close()

        trajectory = np.array([[[trajectory[i][j*number_of_dimensions+k]
                                 for k in range(number_of_dimensions)]
                                for j in range(number_of_atoms)]
                               for i in range (len(trajectory))])

        if last_steps is not None:
            trajectory = trajectory[-last_steps:,:,:]
            energy = energy[-last_steps:]

        print('Number of total steps read: {0}'.format(trajectory.shape[0]))
        time = np.array([i*time_step for i in range(trajectory.shape[0])], dtype=float)

        print('Trajectory file read')
        return dyn.Dynamics(structure=structure,
                            trajectory=np.array(trajectory, dtype=complex),
                            energy=np.array(energy),
                            time=time,
                            supercell=super_cell,
                            memmap=memmap)


# LAMMPS custom dump file parser
def read_lammps_trajectory(file_name, structure=None, time_step=None,
                           limit_number_steps=10000000,
                           last_steps=None,
                           initial_cut=1,
                           end_cut=None,
                           memmap=False,
                           template=None):


    # Time in picoseconds
    # Coordinates in Angstroms

    # Read environtment variables
    try:
        temp_directory = os.environ["DYNAPHOPY_TEMPDIR"]
        if os.path.isdir(temp_directory):
            print('Set temporal directory: {0}'.format(temp_directory))
            temp_directory += '/'
        else:
            temp_directory = ''
    except KeyError:
        temp_directory = ''

    number_of_atoms = None
    bounds = None

    #Check file exists
    if not os.path.isfile(file_name):
        print('Trajectory file does not exist!')
        exit()

    # Check time step
    if time_step is None:
        print('Warning! LAMMPS trajectory file does not contain time step information')
        print('Using default: 0.001 ps')
        time_step = 0.001

    # Starting reading
    print("Reading LAMMPS trajectory")
    print("This may take long, please wait..")

    # Dimension of LAMMP calculation
    if structure is None:
        number_of_dimensions = 3
    else:
        number_of_dimensions = structure.get_number_of_dimensions()

    time = []
    data = []
    counter = 0

    lammps_labels = False

    with open(file_name, "r+") as f:

        file_map = mmap.mmap(f.fileno(), 0)

        while True:

            counter += 1

            #Read time steps
            position_number=file_map.find(b'TIMESTEP')
            if position_number < 0: break

            file_map.seek(position_number)
            file_map.readline()
            time.append(float(file_map.readline()))


            if number_of_atoms is None:
                #Read number of atoms
                file_map = mmap.mmap(f.fileno(), 0)
                position_number=file_map.find(b'NUMBER OF ATOMS')
                file_map.seek(position_number)
                file_map.readline()
                number_of_atoms = int(file_map.readline())

                # Check if number of atoms is multiple of cell atoms
                if structure is not None:
                    if number_of_atoms % structure.get_number_of_cell_atoms() != 0:
                        print('Warning: Number of atoms not matching, check LAMMPS output file')

            if bounds is None:
                #Read cell
                file_map = mmap.mmap(f.fileno(), 0)
                position_number=file_map.find(b'BOX BOUNDS')
                file_map.seek(position_number)
                file_map.readline()

                bounds = []
                for i in range(3):
                    bounds.append(file_map.readline().split())

                bounds = np.array(bounds, dtype=float)
                if bounds.shape[1] == 2:
                    bounds = np.append(bounds, np.array([0, 0, 0])[None].T ,axis=1)

                xy = bounds[0, 2]
                xz = bounds[1, 2]
                yz = bounds[2, 2]

                xlo = bounds[0, 0] - np.min([0.0, xy, xz, xy+xz])
                xhi = bounds[0, 1] - np.max([0.0, xy, xz, xy+xz])
                ylo = bounds[1, 0] - np.min([0.0, yz])
                yhi = bounds[1, 1] - np.max([0.0, yz])
                zlo = bounds[2, 0]
                zhi = bounds[2, 1]

                supercell = np.array([[xhi-xlo, xy,  xz],
                                       [0,  yhi-ylo,  yz],
                                       [0,   0,  zhi-zlo]]).T

                #for 2D
                supercell = supercell[:number_of_dimensions, :number_of_dimensions]

                # Testing cell
                lx = xhi-xlo
                ly = yhi-ylo
                lz = zhi-zlo

                a = lx
                b = np.sqrt(pow(ly,2) + pow(xy,2))
                c = np.sqrt(pow(lz,2) + pow(xz,2) +  pow(yz,2))

                alpha = np.arccos((xy*xz + ly*yz)/(b*c))
                beta = np.arccos(xz/c)
                gamma = np.arccos(xy/b)

                # End testing cell

                # rotate lammps supercell to match unitcell orientation
                def unit_matrix(matrix):
                    return np.array([np.array(row)/np.linalg.norm(row) for row in matrix])

                unit_structure = unit_matrix(structure.get_cell())
                unit_supercell_lammps = unit_matrix(supercell)

                transformation_mat = np.dot(np.linalg.inv(unit_structure), unit_supercell_lammps).T

                supercell = np.dot(supercell, transformation_mat)

                if memmap:
                    if end_cut:
                        data = np.memmap(temp_directory+'trajectory.{0}'.format(os.getpid()), dtype='complex', mode='w+', shape=(end_cut - initial_cut+1, number_of_atoms, number_of_dimensions))
                    else:
                        print('Memory mapping requires to define reading range (use read_from/read_to option)')
                        exit()

            position_number = file_map.find(b'ITEM: ATOMS')

            file_map.seek(position_number)
            lammps_labels=file_map.readline()

            #Initial cut control
            if initial_cut > counter:
                time = []
                continue

            #Reading coordinates
            read_coordinates = []
            for i in range (number_of_atoms):
                read_coordinates.append(file_map.readline().split()[0:number_of_dimensions])
            read_coordinates = np.array(read_coordinates, dtype=float)

            if template is not None:
                indexing = np.argsort(template)
                read_coordinates = read_coordinates[indexing, :]

            try:
                read_coordinates = np.dot(read_coordinates, transformation_mat)
                if memmap:
                    data[counter-initial_cut, :, :] = read_coordinates #in angstroms
                else:
                    data.append(read_coordinates) #in angstroms

            except ValueError:
                print("Error reading step {0}".format(counter))
                break
                # print(read_coordinates)

            #security routine to limit maximum of steps to read and put in memory
            if limit_number_steps+initial_cut < counter:
                print("Warning! maximum number of steps reached! No more steps will be read")
                break

            if end_cut is not None and end_cut <= counter:
                break

    file_map.close()

    time = np.array(time) * time_step

    if not memmap:
        data = np.array(data, dtype=complex)

        if last_steps is not None:
            data = data[-last_steps:, :, :]
            time = time[-last_steps:]


    # Check position/velocity dump
    if b'vx vy' in lammps_labels:
        return dyn.Dynamics(structure=structure,
                            velocity=data,
                            time=time,
                            supercell=supercell,
                            memmap=memmap)

    if b'x y' in lammps_labels:
        return dyn.Dynamics(structure=structure,
                            trajectory=data,
                            time=time,
                            supercell=supercell,
                            memmap=memmap)

    print('LAMMPS parsing error. Data not recognized: {}'.format(lammps_labels))
    exit()


def read_VASP_XDATCAR(file_name, structure=None, time_step=None,
                      limit_number_steps=10000000,
                      last_steps=None,
                      initial_cut=1,
                      end_cut=None,
                      memmap=False,
                      template=None):

    # Time in picoseconds
    # Coordinates in Angstroms

    #Read environtment variables
    try:
        temp_directory = os.environ["DYNAPHOPY_TEMPDIR"]
        if os.path.isdir(temp_directory):
            print('Set temporal directory: {0}'.format(temp_directory))
            temp_directory += '/'
        else:
            temp_directory = ''
    except KeyError:
        temp_directory = ''

    number_of_atoms = None
    bounds = None

    #Check file exists
    if not os.path.isfile(file_name):
        print('Trajectory file does not exist!')
        exit()

    #Check time step
    if time_step is None:
        print('Warning! XDATCAR file does not contain time step information')
        print('Using default: 0.001 ps')
        time_step = 0.001

    #Starting reading
    print("Reading XDATCAR file")
    print("This may take long, please wait..")

    #Dimensionality of VASP calculation
    number_of_dimensions = 3

    time = []
    data = []
    counter = 0

    with open(file_name, "r+b") as f:

        file_map = mmap.mmap(f.fileno(), 0)

        #Read cell
        for i in range(2): file_map.readline()
        a = file_map.readline().split()
        b = file_map.readline().split()
        c = file_map.readline().split()
        super_cell = np.array([a, b, c], dtype='double')

        for i in range(1): file_map.readline()
        number_of_atoms = np.array(file_map.readline().split(), dtype=int).sum()

        while True:

            counter += 1
            #Read time steps
            position_number=file_map.find(b'Direct configuration')
            if position_number < 0: break

            file_map.seek(position_number)
            time.append(float(file_map.readline().split(b'=')[1]))

            if memmap:
                if end_cut:
                    data = np.memmap(temp_directory+'trajectory.{0}'.format(os.getpid()), dtype='complex', mode='w+', shape=(end_cut - initial_cut+1, number_of_atoms, number_of_dimensions))
                else:
                    print('Memory mapping requires to define reading range (use read_from/read_to option)')
                    exit()


            #Initial cut control
            if initial_cut > counter:
                continue

            #Reading coordinates
            read_coordinates = []
            for i in range (number_of_atoms):
                read_coordinates.append(file_map.readline().split()[0:number_of_dimensions])

            read_coordinates = np.array(read_coordinates, dtype=float)  # in angstroms
            if template is not None:
                indexing = np.argsort(template)
                read_coordinates = read_coordinates[indexing, :]

            try:
                if memmap:
                    data[counter-initial_cut, :, :] = read_coordinates #in angstroms
                else:
                    data.append(read_coordinates) #in angstroms

            except ValueError:
                print("Error reading step {0}".format(counter))
                break
                # print(read_coordinates)

            #security routine to limit maximum of steps to read and put in memory
            if limit_number_steps+initial_cut < counter:
                print("Warning! maximum number of steps reached! No more steps will be read")
                break

            if end_cut is not None and end_cut <= counter:
                break

    file_map.close()

    time = np.array(time) * time_step

    if not memmap:
        data = np.array(data, dtype=complex)

        if last_steps is not None:
            data = data[-last_steps:, :, :]
            time = time[-last_steps:]


    return dyn.Dynamics(structure=structure,
                        scaled_trajectory=data,
                        time=time,
                        supercell=super_cell,
                        memmap=memmap)


if __name__ == "__main__":
    read_VASP_XDATCAR('/home/abel/VASP/MgO/MgO-FINAL/MgO_0.5_1600/No1/XDATCAR')



"""Solver of 4th order force constants."""

from __future__ import annotations

import time
from typing import Literal, Optional

import numpy as np

from macer.externals.symfc_bundled.basis_sets import FCBasisSetO4
from macer.externals.symfc_bundled.solvers.solver_O2O3O4 import (
    reshape_nNNN3333_nx_to_N3N3N3_n3nx,
    set_disps_N3N3N3,
)
from macer.externals.symfc_bundled.utils.matrix import block_matrix_sandwich
from macer.externals.symfc_bundled.utils.solver_funcs import get_batch_slice, solve_linear_equation

try:
    from macer.externals.symfc_bundled.utils.matrix import dot_product_sparse
except ImportError:
    pass

from .solver_base import FCSolverBase


class FCSolverO4(FCSolverBase):
    """Fourth order force constants solver."""

    def __init__(
        self,
        basis_set: FCBasisSetO4,
        use_mkl: bool = False,
        log_level: int = 0,
    ):
        """Init method."""
        self._basis_set: FCBasisSetO4
        super().__init__(basis_set, use_mkl=use_mkl, log_level=log_level)

    def solve(
        self,
        displacements: np.ndarray,
        forces: np.ndarray,
        batch_size: int = 36,
    ) -> FCSolverO4:
        """Solve force constants.

        Note
        ----
        self._coefs = coefs_fc4

        Parameters
        ----------
        displacements : ndarray
            Displacements of atoms in Cartesian coordinates.
            shape=(n_snapshot, N, 3), dtype='double'
        forces : ndarray
            Forces of atoms in Cartesian coordinates.
            shape=(n_snapshot, N, 3), dtype='double'

        Returns
        -------
        self : FCSolverO4

        """
        n_data = forces.shape[0]
        f = forces.reshape(n_data, -1)
        d = displacements.reshape(n_data, -1)

        fc4_basis = self._basis_set
        compress_mat_fc4 = fc4_basis.compact_compression_matrix
        basis_set_fc4 = fc4_basis.blocked_basis_set

        atomic_decompr_idx_fc4 = fc4_basis.atomic_decompr_idx

        self._coefs = run_solver_O4(
            d,
            f,
            compress_mat_fc4,
            basis_set_fc4,
            atomic_decompr_idx_fc4,
            batch_size=batch_size,
            use_mkl=self._use_mkl,
            verbose=self._log_level > 0,
        )
        return self

    @property
    def full_fc(self) -> Optional[np.ndarray]:
        """Return full force constants.

        Returns
        -------
        np.ndarray
            shape=(N, N, N, N, 3, 3, 3, 3), dtype='double', order='C'

        """
        return self._recover_fcs("full")

    @property
    def compact_fc(self) -> Optional[np.ndarray]:
        """Return full force constants.

        Returns
        -------
        np.ndarray
            shape=(n_a, N, N, N, 3, 3, 3, 3), dtype='double', order='C'

        """
        return self._recover_fcs("compact")

    def _recover_fcs(
        self, comp_mat_type: Literal["full", "compact"]
    ) -> Optional[np.ndarray]:
        fc4_basis: FCBasisSetO4 = self._basis_set

        if self._coefs is None or fc4_basis.basis_set is None:
            return None

        if comp_mat_type == "full":
            comp_mat_fc4 = fc4_basis.compression_matrix
        elif comp_mat_type == "compact":
            comp_mat_fc4 = fc4_basis.compact_compression_matrix
        else:
            raise ValueError("Invalid comp_mat_type.")

        N = self._natom
        fc4 = fc4_basis.blocked_basis_set.dot(self._coefs)
        fc4 = np.array(
            (comp_mat_fc4 @ fc4).reshape((-1, N, N, N, 3, 3, 3, 3)),
            dtype="double",
            order="C",
        )

        return fc4


def prepare_normal_equation_O4(
    disps,
    forces,
    compact_compress_mat_fc4,
    compress_eigvecs_fc4,
    atomic_decompr_idx_fc4,
    batch_size=36,
    use_mkl=False,
    verbose=False,
):
    r"""Calculate X.T @ X and X.T @ y.

    X = displacements (fc4) @ compress_mat @ compress_eigvecs

    displacements (fc4): (n_samples, NNN333)
    compact_compress_mat_fc4: (n_aNNN3333, n_compr_fc4)
    compress_eigvecs_fc4: (n_compr_fc4, n_basis_fc4)
    Matrix reshapings are appropriately applied to compress_mat
    and its products.

    X.T @ X and X.T @ y are sequentially calculated using divided dataset.
    X.T @ X = \sum_i X_i.T @ X_i
    X.T @ y = \sum_i X_i.T @ y_i (i: batch index)
    """
    N3 = disps.shape[1]
    N = N3 // 3
    NNN = N**3
    n_compr_fc4 = compact_compress_mat_fc4.shape[1]

    n_batch = (n_compr_fc4 // 5000 + 1) * (N // 50 + 1)
    n_batch = min(N, n_batch)
    begin_batch_atom, end_batch_atom = get_batch_slice(N, N // n_batch)
    begin_batch, end_batch = get_batch_slice(disps.shape[0], batch_size)

    mat44 = np.zeros((n_compr_fc4, n_compr_fc4), dtype=float)
    mat4y = np.zeros(n_compr_fc4, dtype=float)

    t_all1 = time.time()
    const_fc4 = -1.0 / 6.0
    compact_compress_mat_fc4 *= const_fc4
    for begin_i, end_i in zip(begin_batch_atom, end_batch_atom, strict=True):
        if verbose:
            print("-----", flush=True)
            print("Solver_atoms:", begin_i + 1, "--", end_i, "/", N, flush=True)
        n_atom_batch = end_i - begin_i

        t1 = time.time()
        decompr_idx = (
            atomic_decompr_idx_fc4[begin_i * NNN : end_i * NNN, None] * 81
            + np.arange(81)[None, :]
        ).reshape(-1)
        compr_mat_fc4 = reshape_nNNN3333_nx_to_N3N3N3_n3nx(
            compact_compress_mat_fc4[decompr_idx],
            N,
            n_atom_batch,
        )
        t2 = time.time()
        if verbose:
            print(
                "Time (Solver_compr_matrix_reshape):",
                "{:.3f}".format(t2 - t1),
                flush=True,
            )

        for begin, end in zip(begin_batch, end_batch, strict=True):
            t1 = time.time()
            X4 = dot_product_sparse(
                set_disps_N3N3N3(disps[begin:end], sparse=False),
                compr_mat_fc4,
                use_mkl=use_mkl,
                dense=True,
            ).reshape((-1, n_compr_fc4))
            y = forces[begin:end, begin_i * 3 : end_i * 3].reshape(-1)
            mat44 += X4.T @ X4
            mat4y += X4.T @ y
            t2 = time.time()
            if verbose:
                print("Solver_block:", end, "/", disps.shape[0], flush=True)
                print(" - Time:", "{:.3f}".format(t2 - t1), flush=True)

    if verbose:
        print("Solver:", "Calculate X.T @ X and X.T @ y", flush=True)

    XTX = block_matrix_sandwich(compress_eigvecs_fc4, compress_eigvecs_fc4, mat44)
    XTy = compress_eigvecs_fc4.transpose_dot(mat4y)

    compact_compress_mat_fc4 /= const_fc4
    t_all2 = time.time()
    if verbose:
        print(
            "Time (disp @ compr @ eigvecs).T @ (disp @ compr @ eigvecs):",
            "{:.3f}".format(t_all2 - t_all1),
            flush=True,
        )
    return XTX, XTy


def run_solver_O4(
    disps,
    forces,
    compact_compress_mat_fc4,
    compress_eigvecs_fc4,
    atomic_decompr_idx_fc4,
    batch_size=36,
    use_mkl=False,
    verbose=False,
):
    """Estimate coeffs. in X @ coeffs = y.

    X = displacements_fc4 @ compress_mat_fc4 @ compress_eigvecs_fc4

    Matrix reshapings are appropriately applied.
    X: features (n_samples * N3, N_basis_fc4)
    y: observations (forces), (n_samples * N3)

    """
    XTX, XTy = prepare_normal_equation_O4(
        disps,
        forces,
        compact_compress_mat_fc4,
        compress_eigvecs_fc4,
        atomic_decompr_idx_fc4,
        batch_size=batch_size,
        use_mkl=use_mkl,
        verbose=verbose,
    )
    coefs = solve_linear_equation(XTX, XTy)
    return coefs



import torch
from macer.externals.e3nn_legacy.math import normalize2mom
from macer.externals.e3nn_legacy.util.jit import compile_mode
from macer.externals.e3nn_legacy.o3 import SO3Grid


@compile_mode('script')
class SO3Activation(torch.nn.Module):
    r'''Apply non linearity on the signal on SO(3)

    Parameters
    ----------
    lmax_in : int
        input lmax

    lmax_out : int
        output lmax

    act : function
        activation function :math:`\phi`

    resolution : int
        SO(3) grid resolution

    normalization : {'norm', 'component'}
    '''
    def __init__(self, lmax_in, lmax_out, act, resolution, *, normalization='component', aspect_ratio=2):
        super().__init__()

        self.grid_in = SO3Grid(lmax_in, resolution, normalization=normalization, aspect_ratio=aspect_ratio)
        self.grid_out = SO3Grid(lmax_out, resolution, normalization=normalization, aspect_ratio=aspect_ratio)
        self.act = normalize2mom(act)

        self.lmax_in = lmax_in
        self.lmax_out = lmax_out

    def __repr__(self):
        return f"{self.__class__.__name__} ({self.lmax_in} -> {self.lmax_out})"

    def forward(self, features):
        r'''evaluate

        Parameters
        ----------

        features : `torch.Tensor`
            tensor of shape ``(..., self.irreps_in.dim)``

        Returns
        -------
        `torch.Tensor`
            tensor of shape ``(..., self.irreps_out.dim)``
        '''
        features = self.grid_in.to_grid(features)
        features = self.act(features)
        features = self.grid_out.from_grid(features)

        return features



"""Utility functions for eigenvalue solutions."""

from collections import defaultdict
from dataclasses import dataclass
from typing import Optional, Union

import numpy as np
import scipy
from scipy.sparse import csr_array

from macer.externals.symfc_bundled.utils.graph import connected_components
from macer.externals.symfc_bundled.utils.matrix import (
    BlockMatrixNode,
    append_node,
    matrix_rank,
    return_numpy_array,
    root_block_matrix,
)
from macer.externals.symfc_bundled.utils.solver_funcs import get_batch_slice


def eigh_projector(
    p: Union[np.ndarray, csr_array],
    atol: float = 1e-8,
    rtol: float = 0.0,
    return_cmplt: bool = False,
    return_block: bool = False,
    verbose: bool = True,
) -> Union[
    np.ndarray,
    tuple[np.ndarray, tuple[np.ndarray, np.ndarray]],
    None,
    tuple[None, None],
]:
    """Solve eigenvalue problem using numpy and eliminate eigenvectors with e < 1.0."""
    p = return_numpy_array(p)
    rank = matrix_rank(p)
    if rank == 0:
        if return_cmplt:
            return None, None
        return None

    if rank > 32767:
        raise RuntimeError("Projector rank is too large in eigh.")

    try:
        eigvals, eigvecs = np.linalg.eigh(p)
    except np.linalg.LinAlgError as e:
        if verbose:
            print(f"np.linalg.eigh failed: {str(e)}")
            print("Try scipy.linalg.lapack.dsyev")
        eigvals, eigvecs, info = scipy.linalg.lapack.dsyev(p.T)
        if info != 0:
            raise scipy.linalg.LinAlgError(
                "scipy.linalg.lapack.dsyev failed: Eigenvalues did not converge"
            ) from e

    tol = 1e-8
    if np.count_nonzero((eigvals > 1.0 + tol) | (eigvals < -tol)):
        diff = np.abs(p - p.T)
        if np.any(diff > 1e-3):
            raise RuntimeError("Transpose equality not satisfied")
        elif np.any(diff > 1e-15):
            eigvals, eigvecs = np.linalg.eigh(0.5 * (p + p.T))

    if np.count_nonzero((eigvals > 1.0 + tol) | (eigvals < -tol)):
        raise ValueError("Eigenvalue error: e > 1 or e < 0.")

    nonzero = np.isclose(eigvals, 1.0, atol=atol, rtol=rtol)
    if return_cmplt:
        cmplt_bool = np.logical_and(np.logical_not(nonzero), eigvals > 1e-12)
        cmplt_eigvals, cmplt_eigvecs = eigvals[cmplt_bool], eigvecs[:, cmplt_bool]

    eigvecs = eigvecs[:, nonzero]
    if return_block:
        block = root_block_matrix(data=eigvecs)

        if return_cmplt:
            return block, (cmplt_eigvals, cmplt_eigvecs)
        return block

    if return_cmplt:
        return eigvecs, (cmplt_eigvals, cmplt_eigvecs)
    return eigvecs


def _compr_projector(p: csr_array) -> tuple[csr_array, Optional[csr_array]]:
    """Compress projection matrix p with many zero rows and columns."""
    _, col_p = p.nonzero()  # type: ignore
    col_p = np.unique(col_p)
    size = len(col_p)

    if p.shape[1] > size:
        compr = csr_array(
            (np.ones(size), (col_p, np.arange(size))),
            shape=(p.shape[1], size),
            dtype="int",
        )
        """p = compr.T @ p @ compr"""
        p = p[col_p].T
        p = p[col_p].T
        return p, compr
    return p, None


def _find_projector_blocks(p: csr_array, verbose: bool = False):
    """Find block structures in projection matrix."""
    if verbose:
        print("Finding block diagonal structure in projector.", flush=True)

    if len(p.data) < 2147483647:
        if verbose:
            print("Using scipy connected_components.", flush=True)
        n_components, labels = scipy.sparse.csgraph.connected_components(p)
        group = defaultdict(list)
        for i, ll in enumerate(labels):
            group[ll].append(i)
    else:
        if verbose:
            print("Using symfc connected_components with DFS.", flush=True)
        group = connected_components(p)

    return group


def _recover_eigvecs_from_uniq_eigvecs(
    uniq_eigvecs: dict,
    group: dict,
    size_projector: int,
) -> csr_array:
    """Recover all eigenvectors from unique eigenvectors.

    Parameters
    ----------
    uniq_eigvecs: Unique eigenvectors and submatrixblock indices
                  are included in its values.
    group: Row indices comprising submatrix blocks.

    """
    total_length = sum(
        len(labels) * v.shape[0] * v.shape[1]
        for v, labels in uniq_eigvecs.values()
        if v is not None
    )
    row = np.zeros(total_length, dtype=int)
    col = np.zeros(total_length, dtype=int)
    data = np.zeros(total_length, dtype="double")

    current_id, col_id = 0, 0
    for eigvecs, labels in uniq_eigvecs.values():
        if eigvecs is not None:
            n_row, n_col = eigvecs.shape
            num_labels = len(labels)
            end_id = current_id + n_row * n_col * num_labels

            row[current_id:end_id] = np.repeat(
                [i for ll in labels for i in group[ll]], n_col
            )
            col[current_id:end_id] = [
                j
                for seq, _ in enumerate(labels)
                for i in range(n_row)
                for j in range(col_id + seq * n_col, col_id + (seq + 1) * n_col)
            ]
            data[current_id:end_id] = np.tile(eigvecs.flatten(), num_labels)

            col_id += n_col * num_labels
            current_id = end_id

    n_col = col_id
    eigvecs = csr_array((data, (row, col)), shape=(size_projector, n_col))
    return eigvecs


@dataclass
class DataCSR:
    """Dataclass for extracting data in projector."""

    data: np.ndarray
    block_labels: np.ndarray
    block_sizes: np.ndarray
    slice_begin: Optional[np.ndarray] = None
    slice_end: Optional[np.ndarray] = None

    def __post_init__(self):
        """Init method."""
        self.slice_end = np.cumsum(self.block_sizes**2)
        self.slice_begin = np.zeros_like(self.slice_end)
        self.slice_begin[1:] = self.slice_end[:-1]

    def get_data(self, idx: int):
        """Get data in projector for i-th block."""
        if self.slice_begin is None or self.slice_end is None:
            raise ValueError("Slice begin and end are not initialized.")
        s1 = self.slice_begin[idx]
        s2 = self.slice_end[idx]
        return self.data[s1:s2]

    def get_block_label(self, idx: int):
        """Get block label for i-th block."""
        return self.block_labels[idx]

    def get_block_size(self, idx: int):
        """Get block size for i-th block."""
        return self.block_sizes[idx]

    @property
    def n_blocks(self):
        """Return number of blocks in projector."""
        return len(self.block_labels)


def _extract_sparse_projector_data(p: csr_array, group: dict) -> DataCSR:
    """Extract data in projector in csr_format efficiently.

    Parameters
    ----------
    p: Projection matrix in CSR format.
    group: Row indices comprising submatrix blocks.

    """
    # r = np.array([i for ids in group.values() for i in ids for j in ids])
    group_ravel = [i for ids in group.values() for i in ids]
    lengths = [len(ids) for ids in group.values() for i in ids]
    r = np.repeat(group_ravel, lengths)
    c = np.array([j for ids in group.values() for _ in ids for j in ids])
    sizes = np.array([len(ids) for ids in group.values()], dtype=int)

    p_data = DataCSR(
        data=np.ravel(p[r, c]),
        block_labels=np.array(list(group.keys())),
        block_sizes=sizes,
    )
    return p_data


def eigsh_projector(
    p: csr_array,
    atol: float = 1e-8,
    rtol: float = 0.0,
    verbose: bool = True,
) -> csr_array:
    """Solve eigenvalue problem for matrix p.

    Return sparse matrix for eigenvectors of matrix p.

    This algorithm begins with finding block diagonal structures in matrix p.
    For each block submatrix, eigenvectors are solved. When p = diag(A,B), Av =
    v, and Bw = w, p[v,0] = [v,0] and p[0,w] = [0,w] are solutions.

    This function avoids solving numpy.eigh for duplicate block matrices. This
    function is efficient for matrix p composed of many duplicate block
    matrices.

    """
    p, compr_p = _compr_projector(p)
    group = _find_projector_blocks(p, verbose=verbose)
    if verbose:
        rank = matrix_rank(p)
        print("Rank of projector:", rank, flush=True)
        print("Number of blocks in projector:", len(group), flush=True)

    p_data = _extract_sparse_projector_data(p, group)
    uniq_eigvecs = dict()
    for i in range(p_data.n_blocks):
        p_block = p_data.get_data(i)
        block_label = p_data.get_block_label(i)
        block_size = p_data.get_block_size(i)
        if block_size > 1:
            key = tuple(p_block)
            try:
                uniq_eigvecs[key][1].append(block_label)
            except KeyError:
                p_np = p_block.reshape((block_size, block_size))
                eigvecs = eigh_projector(p_np, atol=atol, rtol=rtol, verbose=verbose)
                uniq_eigvecs[key] = [eigvecs, [block_label]]
        else:
            if not np.isclose(p_block[0], 0.0):
                if "one" in uniq_eigvecs:
                    uniq_eigvecs["one"][1].append(block_label)
                else:
                    uniq_eigvecs["one"] = [np.array([[1.0]]), [block_label]]

    c_p = _recover_eigvecs_from_uniq_eigvecs(uniq_eigvecs, group, p.shape[0])  # type: ignore
    if compr_p is not None:
        return compr_p @ c_p
    return c_p


def _find_submatrix_eigenvectors(
    p: Union[np.ndarray, csr_array],
    batch_size: Optional[int] = None,
    depth: int = 0,
    use_mkl: bool = False,
    verbose: bool = False,
):
    """Find eigenvectors in division part of submatrix division algorithm."""
    p_size = p.shape[0]
    if p_size < 500:
        repeat = False
    elif p_size > 30000:
        repeat = True if depth < 8 else False
    else:
        repeat = True if depth < 3 else False

    if batch_size is None:
        if depth == 1:
            batch_size = max(p_size // 10, 500)
        elif depth == 2:
            batch_size = max(p_size // 5, 250)
        elif depth == 3:
            batch_size = p_size // 2
        else:
            batch_size = min(int(round(p_size / 1.3)), 20000)

    sibling, sibling_c = None, None
    col_id, col_id_c = 0, 0
    header = "  " * (depth - 1) + "(" + str(depth) + ")"
    for begin, end in zip(*get_batch_slice(p_size, batch_size), strict=True):
        if verbose:
            print(header, "Block:", end, "/", p_size, flush=True)
        p_small = p[begin:end, begin:end]
        rank = matrix_rank(p_small)
        if rank > 0:
            if repeat:
                block, (cmplt_eigvals, cmplt_vecs) = eigh_projector_division(
                    p_small,
                    atol=1e-12,
                    rtol=0.0,
                    depth=depth,
                    return_cmplt=True,
                    return_block=True,
                    use_mkl=use_mkl,
                    verbose=verbose,
                )
            else:
                block, (cmplt_eigvals, cmplt_vecs) = eigh_projector(
                    p_small,
                    atol=1e-12,
                    rtol=0.0,
                    return_cmplt=True,
                    return_block=True,
                    verbose=verbose,
                )

            rows = np.arange(begin, end)
            if block is not None:
                if verbose:
                    print(
                        header, " ", block.shape[1], "eigenvectors found.", flush=True
                    )

                sibling = append_node(block, sibling, rows=rows, col_begin=col_id)
                col_id += block.shape[1]
            if cmplt_vecs is not None:
                sibling_c = append_node(
                    cmplt_vecs,
                    sibling_c,
                    rows=rows,
                    col_begin=col_id_c,
                    eigvals=cmplt_eigvals,
                )
                col_id_c += cmplt_vecs.shape[1]

    if col_id_c > 0:
        cmplt = root_block_matrix((p_size, col_id_c), first_child=sibling_c)
    else:
        cmplt = None
    return sibling, col_id, cmplt


def _find_complement_eigenvectors(
    p: Union[np.ndarray, csr_array],
    sibling: BlockMatrixNode,
    col_id: int,
    cmplt: BlockMatrixNode,
    atol: float = 1e-8,
    rtol: float = 0.0,
    depth: int = 0,
    return_cmplt: bool = False,
    use_mkl: bool = False,
    verbose: bool = False,
):
    """Find eigenvectors in complementary part of submatrix division algorithm."""
    p_size = p.shape[0]
    if p_size < 5000:
        repeat = False
    elif p_size > 30000:
        repeat = True if depth < 8 else False
    else:
        repeat = True if depth < 4 else False

    if depth == 1:
        batch_size_cmplt = max(cmplt.shape[1] // 3, p_size // 15)
    elif depth == 2:
        batch_size_cmplt = int(round(cmplt.shape[1] / 1.5))
    elif depth == 3:
        batch_size_cmplt = int(round(cmplt.shape[1] / 1.3))
    else:
        batch_size_cmplt = min(int(round(cmplt.shape[1] / 1.2)), 20000)

    header = "  " * (depth - 1) + "(" + str(depth) + ")"
    if verbose:
        print(header, "Complementary block size:", cmplt.shape[1], flush=True)
        print(header, "Compute compressed projector.", flush=True)

    p_cmr = cmplt.compress_matrix(p, use_mkl=use_mkl)
    if not repeat:
        if verbose:
            print(header, "Use standard solver.", flush=True)
        result = eigh_projector(
            p_cmr,
            atol=atol,
            rtol=rtol,
            return_cmplt=return_cmplt,
            verbose=verbose,
        )
    else:
        if verbose:
            print(header, "Use submatrix size of", batch_size_cmplt, flush=True)
        result = eigh_projector_division(
            p_cmr,
            atol=atol,
            rtol=rtol,
            depth=depth,
            batch_size=batch_size_cmplt,
            return_cmplt=return_cmplt,
            use_mkl=use_mkl,
            verbose=verbose,
        )

    eigvecs = result[0] if return_cmplt else result

    if eigvecs is not None:
        if verbose:
            print(header, " ", eigvecs.shape[1], "eigenvectors found.", flush=True)
        sibling = append_node(
            eigvecs,
            sibling,
            rows=np.arange(p_size),
            col_begin=col_id,
            compress=cmplt,
        )
        col_id += eigvecs.shape[1]

    if return_cmplt:
        cmplt_eigvals, cmplt_small = result[1]
        if cmplt_small is not None and cmplt_small.shape[1] > 0:
            cmplt_eigvecs = cmplt.dot(cmplt_small)
        else:
            cmplt_eigvals, cmplt_eigvecs = None, None
        return sibling, col_id, (cmplt_eigvals, cmplt_eigvecs)
    return sibling, col_id


def _run_division(
    p: Union[np.ndarray, csr_array],
    batch_size: Optional[int] = None,
    atol: float = 1e-8,
    rtol: float = 0.0,
    depth: int = 0,
    return_cmplt: bool = False,
    use_mkl: bool = False,
    verbose: bool = False,
):
    """Find eigenvectors in division and complementary parts."""
    depth += 1
    sibling, col_id, cmplt = _find_submatrix_eigenvectors(
        p,
        batch_size=batch_size,
        depth=depth,
        use_mkl=use_mkl,
        verbose=verbose,
    )
    if cmplt is not None:
        result = _find_complement_eigenvectors(
            p,
            sibling,
            col_id,
            cmplt,
            atol=atol,
            rtol=rtol,
            depth=depth,
            return_cmplt=return_cmplt,
            use_mkl=use_mkl,
            verbose=verbose,
        )
        if return_cmplt:
            sibling, col_id, (cmplt_eigvals, cmplt_eigvecs) = result
        else:
            sibling, col_id = result
    else:
        cmplt_eigvals, cmplt_eigvecs = None, None

    block = root_block_matrix((p.shape[0], col_id), first_child=sibling)
    if return_cmplt:
        return block, (cmplt_eigvals, cmplt_eigvecs)
    return block


def eigh_projector_division(
    p: Union[np.ndarray, csr_array],
    atol: float = 1e-8,
    rtol: float = 0.0,
    depth: int = 0,
    batch_size: Optional[int] = None,
    return_cmplt: bool = False,
    return_block: bool = False,
    use_mkl: bool = False,
    verbose: bool = False,
):
    r"""Solve eigenvalue problem using submatrix division algorithm.

    This algorithm is optimized to solve an eigenvalue problem for
    projection matrix p with large block submatrices.
    The algorithm for solving each block submatrix is as follows.
    1. Divide each block submatrix into reasonable sizes of submatrices,
       not projectors.
    2. Solve eigenvalue problems for these submatrices and eigenvectors
       with eigenvalues of one are extracted. The eigenvectors with
       eigenvalues e < 1 are used for compressing the complementary matrix
       in the next step.
    3. Calculate the complementary matrix, which corresponds to the complement
       of the vector space spanned by the eigenvectors. The complementary matrix
       is compressed using the eigenvectors with e < 1.
    4. Solve eigenvalue problem for the complementary matrix and eigenvectors
       with e = 1 are calculated. The eigenvalue problems are efficiently solved
       using the compressed complementary matrix and the eigenvectors are recovered by
       the compression matrix.
    5. Collect all eigenvectors with e = 1.
    """
    p_size = p.shape[0]
    if p_size < 500:
        return eigh_projector(
            p,
            atol=atol,
            rtol=rtol,
            return_cmplt=return_cmplt,
            return_block=return_block,
            verbose=verbose,
        )

    result = _run_division(
        p,
        batch_size=batch_size,
        atol=atol,
        rtol=rtol,
        depth=depth,
        return_cmplt=return_cmplt,
        use_mkl=use_mkl,
        verbose=verbose,
    )
    if return_block:
        return result

    if return_cmplt:
        block, (cmplt_eigvals, cmplt_eigvecs) = result
    else:
        block = result

    eigvecs = block.recover() if block is not None else None
    if return_cmplt:
        return eigvecs, (cmplt_eigvals, cmplt_eigvecs)
    return eigvecs


def eigsh_projector_sumrule(
    p: csr_array,
    atol: float = 1e-8,
    rtol: float = 0.0,
    size_threshold: int = 500,
    use_mkl: bool = False,
    verbose: bool = True,
) -> BlockMatrixNode:
    """Solve eigenvalue problem for matrix p.

    Return dense matrix for eigenvectors of matrix p.

    This algorithm begins with finding block diagonal structures in matrix p.
    For each block submatrix, eigenvectors are solved.
    When p = diag(A,B), Av = v, and Bw = w, p[v,0] = [v,0] and p[0,w] = [0,w]
    are solutions.

    If p.shape[0] < size_threshold, this function solves numpy.eigh for all block
    matrices. Otherwise, this function use a submatrix division algorithm
    to solve the eigenvalue problem of each block matrix.
    """
    group = _find_projector_blocks(p, verbose=verbose)
    lengths = [-len(ids) for ids in group.values()]
    order = np.array(list(group.keys()))[np.argsort(lengths)]
    if verbose:
        print("Number of blocks in projector (Sum rule):", len(group), flush=True)

    sibling, col_id = None, 0
    for i, key in enumerate(order):
        ids = np.array(group[key])
        if verbose and len(ids) > 0:
            print("--- Eigsh_solver_block:", i + 1, "/", len(group), "---", flush=True)
            print("Block_size:", len(ids), flush=True)

        p_block = p[np.ix_(ids, ids)]
        rank = matrix_rank(p_block)
        if rank > 0 and p_block.shape[0] < size_threshold:
            if verbose:
                print("Use standard eigh solver.", flush=True)
            p_block = p_block.toarray()
            eigvecs = eigh_projector(p_block, atol=atol, rtol=rtol, verbose=verbose)
            if eigvecs is not None:
                sibling = append_node(eigvecs, sibling, rows=ids, col_begin=col_id)
                col_id += eigvecs.shape[1]

        elif rank > 0 and p_block.shape[0] >= size_threshold:
            if verbose:
                print("Use submatrix version of eigh solver.", flush=True)
            block = eigh_projector_division(
                p_block,
                atol=atol,
                rtol=rtol,
                return_block=True,
                use_mkl=use_mkl,
                verbose=verbose,
            )
            if block is not None:
                sibling = append_node(block, sibling, rows=ids, col_begin=col_id)
                col_id += block.shape[1]

        del p_block

    block = root_block_matrix((p.shape[0], col_id), first_child=sibling)
    if verbose:
        print("Tree of FC basis block matrices:", flush=True)
        block.print_nodes()
    return block



import sys
import curses
import numpy as np
import phonopy.file_IO as file_IO
import textwrap
import macer.externals.dynaphopy_bundled as dynaphopy

from time import sleep
from fractions import Fraction
from os.path import isfile


def list_on_screen(screen, pile, posx, posy):

    pile = np.array(pile).reshape((-1,3))

    for i, row_list in enumerate(pile):
        for j, element in enumerate(row_list):
            screen.addstr(posx+i,posy+j*20, str(i*len(pile[0])+j+1)+": {0:.4f}".format(element))


# Get parametres from text ui
def get_param(screen, prompt_string):

    screen_width = 80
    prompt_string = textwrap.fill(prompt_string, width=screen_width)

    screen.clear()
    screen.border(0)

    for i, line in enumerate(prompt_string.splitlines()):
        screen.addstr(2+i, 2, line)
    screen.refresh()
    input_data = screen.getstr(10, 10, 60).decode()

    return input_data


def interactive_interface(calculation, trajectory, args, structure_file):

    screen = curses.initscr()
    screen.border(0)
    screen.addstr(5, 7, "Welcome   to   DynaPhoPy " + dynaphopy.__version__)
    screen.refresh()
    sleep(3)
    curses.endwin()

    x = 0
    while x != ord('0'):
        screen = curses.initscr()
        screen.clear()
        screen.border(0)

        #Show parameters right screen
        screen.addstr(2,45,"Input file: " + args.input_file[0][-20:])
        if args.load_data:
            screen.addstr(4,45,"hdf5 file: " + args.load_data[0][-20:])
        else:
            screen.addstr(3,45,"Structure file: " + structure_file[-14:])
            if args.md_file:
                screen.addstr(4,45,"MD file: " + args.md_file[-20:])
            else:
                screen.addstr(4,45, "Generated trajectory")

        screen.addstr(6,45,"Wave vector: {0} ".format(calculation.get_reduced_q_vector()))
        screen.addstr(7,45,"Frequency range: {0} - {1} THz".format(calculation.get_frequency_range()[0],
                                                                   calculation.get_frequency_range()[-1]))
        screen.addstr(8,45,"Pow. spectr. resolution: {0} THz".format(calculation.parameters.spectrum_resolution))

        screen.addstr(10,45,"Primitive cell atoms: {0}".format(trajectory.structure.get_number_of_primitive_atoms()))
        screen.addstr(11,45,"Unit cell atoms: {0}".format(trajectory.structure.get_number_of_atoms()))
        screen.addstr(12,45,"MD supercell atoms: {0} ".format(trajectory.get_number_of_atoms()))
        screen.addstr(13,45,"Number of MD time steps: {0}".format(len(trajectory.velocity)))
        screen.addstr(14,45,"Time step: {0} ps".format(np.round(trajectory.get_time_step_average(),decimals=12)))


        #Option values left screen
        screen.addstr(2, 2, "Please enter option number...")
        screen.addstr(4, 4, "1 - Harmonic calculations")
        screen.addstr(5, 4, "2 - Change wave vector")
        screen.addstr(6, 4, "3 - Thermal properties")
        screen.addstr(7, 4, "4 - Maxwell-Boltzmann analysis")
        screen.addstr(8, 4, "5 - Power spectrum")
        screen.addstr(9, 4, "6 - Renormalized phonon dispersion")
        screen.addstr(10, 4, "7 - Peak analysis")
        screen.addstr(11, 4, "8 - Atomic displacements distribution")
        screen.addstr(12, 4, "9 - Preferences")
        screen.addstr(14, 4, "0 - Exit")

        screen.refresh()

        x = screen.getch()

######## OPTION 1 :  DISPLAY HARMONIC DATA
        if x == ord('1'):
            x2 = 0
            while x2 != ord('0'):
                screen = curses.initscr()
                screen.clear()
                screen.border(0)

                screen.addstr(2, 2, "Display...")
                screen.addstr(4, 4, "1 - Show harmonic frequencies")
                screen.addstr(5, 4, "2 - Show harmonic eigenvectors")
                screen.addstr(6, 4, "3 - Plot phonon dispersion relations")
                screen.addstr(7, 4, "4 - Plot phonon density of states")
                screen.addstr(9, 4, "0 - Return")
                screen.refresh()

                x2 = screen.getch()

                curses.endwin()

                if x2 == ord('1'):
                    freq = calculation.get_frequencies()
                    sleep(1)
                    screen = curses.initscr()
                    screen.clear()
                    screen.border()

                    screen.addstr(2, 4, "Frequencies (THz)")
                    screen.addstr(3, 4, "-----------------")

                    list_on_screen(screen, freq, 5, 4)
                    screen.getch()

                if x2 == ord('2'):
                    calculation.plot_eigenvectors()

                if x2 == ord('3'):
                    calculation.plot_phonon_dispersion_bands()

                if x2 == ord('4'):
                    calculation.plot_dos_phonopy()

######## OPTION 2 :  DEFINE WAVE VECTOR
        if x == ord('2'):
            q_vector = np.array([float(Fraction(s)) for s in
                                 get_param(screen, "Insert reduced wave vector (values separated by comma)").split(',')])
            calculation.set_reduced_q_vector(q_vector)

######## OPTION 3 :  THERMAL PROPERTIES
        if x == ord('3'):
            curses.endwin()
            calculation.display_thermal_properties()

######## OPTION 4 :  BOLTZMANN DISTRIBUTION
        if x == ord('4'):
            curses.endwin()
            calculation.show_boltzmann_distribution()

######## OPTION 5 :  PLOTTING POWER SPECTRA
        if x == ord('5'):

            x2 = 0
            while x2 != ord('0'):
                screen = curses.initscr()
                screen.clear()
                screen.border(0)

                screen.addstr(2, 2, "Plotting...")
                screen.addstr(4, 4, "1 - Full power spectrum")
                screen.addstr(5, 4, "2 - Wave vector projection power spectrum")
                screen.addstr(6, 4, "3 - Phonon mode projection power spectrum")
                screen.addstr(8, 4, "0 - Return")
                screen.refresh()

                x2 = screen.getch()
                curses.endwin()

                if x2 == ord('1'):
                    calculation.plot_power_spectrum_full()

                if x2 == ord('2'):
                    calculation.plot_power_spectrum_wave_vector()

                if x2 == ord('3'):
                    calculation.plot_power_spectrum_phonon()

######## OPTION 6 :  Renormalized phonon dispersion
        if x == ord('6'):
            x2 = 0
            while x2 != ord('0'):
                screen = curses.initscr()
                screen.clear()
                screen.border(0)

                screen.addstr(2, 2, "Plotting...")
                screen.addstr(4, 4, "1 - Harmonic and renormalized phonon dispersion relations")
                screen.addstr(5, 4, "2 - Renormalized phonon dispersions and linewidths (fat bands)")
                screen.addstr(6, 4, "3 - Frequency shifts and linewidths (separated)")
                screen.addstr(7, 4, "4 - Frequency vs linewidth (interpolated mesh)")
                screen.addstr(9, 4, "0 - Return")
                screen.refresh()

                x2 = screen.getch()

                curses.endwin()

                if x2 == ord('1'):
                    calculation.plot_renormalized_phonon_dispersion_bands()

                if x2 == ord('2'):
                    calculation.plot_renormalized_phonon_dispersion_bands(plot_linewidths=True,
                                                                          plot_harmonic=False)

                if x2 == ord('3'):
                    calculation.plot_linewidths_and_shifts_bands()

                if x2 == ord('4'):
                    calculation.plot_frequencies_vs_linewidths()


######## OPTION 7 :  PEAK ANALYSIS
        if x == ord('7'):
            curses.endwin()
            calculation.phonon_individual_analysis()
            screen.getch()

######## OPTION 8 :  TRAJECTORY DISTRIBUTION
        if x == ord('8'):
            direction = np.array([float(Fraction(s)) for s in
                                 get_param(screen, "Introduce the vector that defines direction in real space (x,y,z) "
                                                   "in which the distribution will be calculated (values separated by comma)").split(',')])
            curses.endwin()
            calculation.plot_trajectory_distribution(direction)

        ######## OPTION 9 :  PREFERENCES
        if x == ord('9'):

            x2 = 0
            while x2 != ord('0'):
                screen = curses.initscr()
                screen.clear()
                screen.border(0)

                screen.addstr(2, 2, "Preferences...")
                screen.addstr(4, 4, "1 - Select power spectrum algorithm")
                screen.addstr(5, 4, "2 - Non analytical corrections (dispersion spectrum only): {0}".format(
                        calculation.parameters.use_NAC))
                screen.addstr(6, 4, "3 - Number of MEM coefficients: {0}".format(
                        calculation.parameters.number_of_coefficients_mem))
                screen.addstr(7, 4, "4 - Number of bins in histograms (Boltzmann/displacements): {0}".format(
                        calculation.parameters.number_of_bins_histogram))
                screen.addstr(8, 4, "5 - Eigenvectors display vector scale: {0}".format(
                        calculation.parameters.modes_vectors_scale))
                screen.addstr(9, 4, "6 - Select fitting function")
                screen.addstr(10, 4, "7 - Change spectrum resolution")
                screen.addstr(11, 4, "8 - Change frequency range")

                screen.addstr(13, 4, "0 - Return")
                screen.refresh()

                x2 = screen.getch()
                # SUB OPT 1 : # PS select algorithm
                if x2 == ord('1'):
                    x3 = 9
                    while x3 >= len(calculation.get_algorithm_list()):
                        screen = curses.initscr()
                        screen.clear()
                        screen.border(0)

                        screen.addstr(2, 2, "Algorithms...")
                        for i, algorithm in enumerate(calculation.get_algorithm_list()):
                            if i == calculation.parameters.power_spectra_algorithm:
                                screen.addstr(4+i, 3, ">"+str(i) +" : "+ str(algorithm[1]))
                            else:
                                screen.addstr(4+i, 4, str(i) +" : "+ str(algorithm[1]))

                        screen.refresh()
                        try:
                            x3 = int(chr(int(screen.getch())))
                        except ValueError:
                            x3 = 9

                    calculation.select_power_spectra_algorithm(x3)

                    curses.endwin()

                # SUB OPT 2 : # Non analytical corrections
                if x2 == ord('2'):
                    x3 = ord("9")
                    while int(chr(int(x3))) > 2:
                        screen = curses.initscr()
                        screen.clear()
                        screen.border(0)

                        screen.addstr(2, 2, "Non analytical corrections...")
                        if isfile("BORN"):
                            screen.addstr(4, 4, "1 - On  (BORN file found)")
                        else:
                            screen.addstr(4, 4, "1 - On  (Warning: BORN file not found)")
                        screen.addstr(5, 4, "2 - Off")


                        if calculation.parameters.use_NAC:
                            screen.addstr(4, 3, ">")
                        else:
                            screen.addstr(5, 3, ">")


                        screen.refresh()
                        x3 = screen.getch()
                    if isfile("BORN"):
                        calculation.set_NAC(bool(int(chr(int(x3)))-2))

                    curses.endwin()

                # SUB OPT 3 : # Number of MEM coefficients
                if x2 == ord('3'):
               # calculation.set_number_of_mem_coefficients(int(get_param(screen, "Insert number of coefficients")))
                    calculation.parameters.number_of_coefficients_mem = int(get_param(screen, "Insert number of coefficients"))
                    curses.endwin()

                # SUB OPT 4 : # Number of bins histogram (coordinates and Boltzmann dist.)
                if x2 == ord('4'):
                    calculation.parameters.number_of_bins_histogram = int(get_param(screen, "Insert number of bins"))
                    curses.endwin()

                # SUB OPT 5 : # Vector scale (phonon modes)
                if x2 == ord('5'):
                    calculation.parameters.modes_vectors_scale = int(get_param(screen, "Insert vector scale"))
                    curses.endwin()

                # SUB OPT 6 : # Fitting function selection
                if x2 == ord('6'):
                    from macer.externals.dynaphopy_bundled.analysis.fitting.fitting_functions import fitting_functions
                    x3 = 9
                    while x3 >= len(fitting_functions):
                        screen = curses.initscr()
                        screen.clear()
                        screen.border(0)

                        screen.addstr(2, 2, "Fitting functions...")
                        for i in fitting_functions.keys():
                            algorithm = fitting_functions[i]
                            if i == calculation.parameters.fitting_function:
                                screen.addstr(4+i, 3, ">"+str(i) +" : "+ str(algorithm).split('.')[-1].replace('_',' '))
                            else:
                                screen.addstr(4+i, 4, str(i) +" : "+ str(algorithm).split('.')[-1].replace('_',' '))

                        screen.refresh()
                        try:
                            x3 = int(chr(int(screen.getch())))
                        except ValueError:
                            x3 = 9

                    calculation.select_fitting_function(x3)

                    curses.endwin()

                # SUB OPT 7 : # Resolution
                if x2 == ord('7'):
                    resolution =float(get_param(screen, "Insert resolution in THz"))
                    calculation.set_spectra_resolution(resolution)
                    calculation.full_clear()
                    curses.endwin()

                # SUB OPT 8 : # Frequency range
                if x2 == ord('8'):
                    frequency_limits = np.array([float(Fraction(s)) for s in
                                                 get_param(screen, "Insert frequency range (min, max)").split(',')])
                    print(frequency_limits)
                    calculation.set_frequency_limits(frequency_limits)
                    curses.endwin()

    curses.endwin()

# Testing
if __name__ == 'test_gui.py':
    import macer.externals.dynaphopy_bundled as controller
    import macer.externals.dynaphopy_bundled.functions.iofunctions as reading

    # Get data from input file
    input_parameters = reading.read_parameters_from_input_file(sys.argv[1])

    structure = reading.read_from_file_structure_outcar(input_parameters['structure_file_name'])
    structure.set_force_set( file_IO.parse_FORCE_SETS(filename=input_parameters['force_constants_file_name']))
    structure.set_primitive_matrix(input_parameters['primitive_matrix'])
    structure.set_supercell_phonon(input_parameters['supercell_matrix'])

    trajectory_file_name = sys.argv[2]

    trajectory = reading.read_from_file_trajectory(trajectory_file_name,structure,last_steps=5000)

    calculation = controller.Quasiparticle(trajectory)
    calculation.set_band_ranges(input_parameters['bands'])


"""Functions for introducing translational invariance in 1st order force constants."""

import numpy as np
import scipy
from scipy.sparse import csr_array


def compressed_projector_sum_rules(
    compress_mat: csr_array, N: int, use_mkl: bool = False
):
    """Return projection matrix for sum rule compressed by C."""
    proj_cplmt = _compressed_complement_projector_sum_rules(
        compress_mat, N, use_mkl=use_mkl
    )
    return scipy.sparse.identity(proj_cplmt.shape[0]) - proj_cplmt


def _compressed_complement_projector_sum_rules_algo1(
    compress_mat: csr_array, N: int, use_mkl: bool = False
):
    r"""Return complementary projection matrix for sum rule compressed by C.

    proj_sum_cplmt = [C.T @ Csum(c)] @ [Csum(c).T @ C]
                   = c_sum_cplmt_compr.T @ c_sum_cplmt_compr
    Matrix shape of proj_sum_cplmt is (C.shape[1], C.shape[1]).
    C.shape[0] must be equal to NN33.

    Sum rules are given as sums over i: \sum_i \phi_{ia,jb} = 0

    """
    N3 = 3 * N

    row = np.arange(N3)
    col = np.tile(range(3), N)
    data = np.zeros(N3)
    data[:] = 1 / np.sqrt(N)
    c_sum_cplmt = csr_array((data, (row, col)), shape=(N3, 3))

    c_sum_cplmt_compr = c_sum_cplmt.T @ compress_mat
    proj_sum_cplmt = c_sum_cplmt_compr.T @ c_sum_cplmt_compr
    return proj_sum_cplmt


def _compressed_complement_projector_sum_rules(
    compress_mat: csr_array, N: int, use_mkl: bool = False
):
    """Return complementary projection matrix for sum rule compressed by C."""
    return _compressed_complement_projector_sum_rules_algo1(
        compress_mat, N, use_mkl=use_mkl
    )



#!/usr/bin/env python3
from __future__ import annotations

import argparse
import subprocess
import sys
from pathlib import Path


def parse_parameters(path: Path) -> dict[str, str]:
    data: dict[str, str] = {}
    for line in path.read_text().splitlines():
        line = line.strip()
        if not line or line.startswith("#") or "=" not in line:
            continue
        k, v = line.split("=", 1)
        data[k.strip()] = v.strip()
    return data


def map_to_macer_cli(params: dict[str, str]) -> list[str]:
    mapped = []
    if "T_K" in params:
        mapped += ["-T", params["T_K"]]
    if "nconf" in params:
        mapped += ["--nconf", params["nconf"]]
    if "nfits" in params:
        mapped += ["--nfits", params["nfits"]]
    if "memory" in params:
        mapped += ["--memory", params["memory"]]
    if "mixing" in params:
        mapped += ["--mixing", params["mixing"]]
    if "tolerance" in params:
        mapped += ["--tolerance", params["tolerance"]]
    if "pdiff" in params:
        mapped += ["--pdiff", params["pdiff"]]
    if "grid" in params:
        mapped += ["--grid", params["grid"]]
    if "imaginary_freq" in params:
        mapped += ["--imaginary-freq", params["imaginary_freq"]]
    if params.get("use_smalldisp", "False").lower() in ("true", "1", "yes"):
        mapped += ["--use-smalldisp"]
    if "use_pressure" in params:
        mapped += ["--use-pressure", params["use_pressure"]]
    if "pressure_diag" in params:
        vals = [x.strip() for x in params["pressure_diag"].split(",")]
        if len(vals) == 3:
            mapped += ["--pressure-diag", vals[0], vals[1], vals[2]]
    return mapped


def _find_poscar(case_dir: Path) -> Path | None:
    for cand in [case_dir / "POSCAR", case_dir / "POSCAR_input"]:
        if cand.exists():
            return cand
    return None


def _override_for_fast(mapped: list[str], fast: bool) -> list[str]:
    if not fast:
        return mapped

    # Remove existing conflicting options first.
    remove_keys = {"--nconf", "--nfits", "--dim", "--no-plot-bands", "--initial-isif"}
    out: list[str] = []
    i = 0
    while i < len(mapped):
        tok = mapped[i]
        if tok in remove_keys:
            i += 2
            continue
        if tok == "--pressure-diag":
            out.extend(mapped[i:i + 4])
            i += 4
            continue
        out.append(tok)
        i += 1

    out += ["--nconf", "1", "--nfits", "1", "--dim", "1", "1", "1", "--initial-isif", "0", "--no-plot-bands"]
    return out


def run_case(case_dir: Path, mapped: list[str], ff: str, timeout_s: int, dry_run: bool) -> tuple[int, str, str]:
    poscar = _find_poscar(case_dir)
    if poscar is None:
        return 2, "", f"POSCAR not found in {case_dir}"

    cmd = [
        sys.executable,
        "-m",
        "macer.cli.phonopy_main",
        "qscaild",
        "-p",
        str(poscar),
        "--ff",
        ff,
    ] + mapped
    if dry_run and "--dry-run" not in cmd:
        cmd.append("--dry-run")

    proc = subprocess.run(
        cmd,
        text=True,
        capture_output=True,
        check=False,
        timeout=timeout_s,
    )
    return proc.returncode, proc.stdout, proc.stderr


def main() -> int:
    p = argparse.ArgumentParser(description="Run smoke tests for qscaild-master reference cases via macer qscaild.")
    p.add_argument("--root", default="qscaild-master/test_dir", help="Reference root containing case directories.")
    p.add_argument("--ff", default="mattersim", help="Force field to use for smoke runs.")
    p.add_argument("--timeout", type=int, default=300, help="Per-case timeout in seconds.")
    p.add_argument("--dry-run", action="store_true", help="Use qscaild --dry-run mode only.")
    p.add_argument("--fast", action="store_true", help="Override to minimal runtime settings.")
    args = p.parse_args()

    root = Path(args.root).resolve()
    param_files = sorted(root.rglob("parameters"))
    if not param_files:
        print(f"No parameters files found under: {root}")
        return 1

    summary: list[tuple[str, int, str]] = []
    for f in param_files:
        case_dir = f.parent
        params = parse_parameters(f)
        mapped = map_to_macer_cli(params)
        mapped = _override_for_fast(mapped, args.fast)
        rc, out, err = run_case(case_dir, mapped, ff=args.ff, timeout_s=args.timeout, dry_run=args.dry_run)
        rel = str(f.relative_to(root))
        if rc == 0:
            msg = "OK"
        else:
            last = (err.strip().splitlines()[-1] if err.strip() else out.strip().splitlines()[-1] if out.strip() else "unknown error")
            msg = f"FAIL: {last}"
        summary.append((rel, rc, msg))
        print(f"[{rel}] {msg}")

    failed = [x for x in summary if x[1] != 0]
    print("")
    print("Summary:")
    print(f"  total={len(summary)} passed={len(summary)-len(failed)} failed={len(failed)}")
    if failed:
        for rel, _, msg in failed:
            print(f"  - {rel}: {msg}")
        return 1
    return 0


if __name__ == "__main__":
    raise SystemExit(main())
